from __future__ import annotations

import asyncio
import contextlib
import time

import pytest

from cashews import Cache
from cashews.backends.memory import Memory

pytestmark = [pytest.mark.asyncio, pytest.mark.redis]


@pytest.fixture(name="create_cache")
def _create_cache(redis_dsn, backend_factory):
    from cashews.backends.redis.client_side import BcastClientSide

    async def call(local_cache=None):
        backend = backend_factory(BcastClientSide, redis_dsn, local_cache=local_cache)
        await backend.init()
        await backend.clear()
        return backend

    return call


async def test_set_get_bcast(create_cache):
    cachef_local = Memory()
    cachef = await create_cache(cachef_local)
    caches_local = Memory()
    caches = await create_cache(caches_local)

    await cachef.set("cashews", b"value", expire=0.1)
    await asyncio.sleep(0.01)  # skip init signal about invalidation
    assert await cachef.get("cashews") == b"value"
    assert await caches.get("cashews") == b"value"
    assert await cachef_local.get("cashews") == b"value"
    assert await caches_local.get("cashews") == b"value"
    await asyncio.sleep(0.2)
    assert await cachef_local.get("cashews") is None
    assert await caches_local.get("cashews") is None

    assert await caches.get("cashews") is None

    await caches.close()
    await cachef.close()


async def test_set_none_bcast(create_cache):
    cachef_local = Memory()
    cachef = await create_cache(cachef_local)
    caches_local = Memory()
    caches = await create_cache(caches_local)

    assert await caches.get("key") is None
    assert not await caches.exists("key")
    assert await caches_local.exists("key")
    assert await caches.get("key") is None

    await cachef.set("key", "val", expire=10000)
    await asyncio.sleep(0.01)  # skip init signal about invalidation
    assert await cachef.exists("key")
    assert await cachef_local.exists("key")

    assert await caches.get("key") == "val"
    assert await caches.exists("key")
    assert await caches_local.exists("key")

    await caches.close()
    await cachef.close()


async def test_del_bcast(create_cache):
    cachef_local = Memory()
    cachef = await create_cache(cachef_local)
    caches_local = Memory()
    caches = await create_cache(caches_local)

    await cachef.set("key", b"value")
    await asyncio.sleep(0.05)  # skip init signal about invalidation

    assert await cachef.get("key") == b"value"
    assert await caches.get("key") == b"value"
    await cachef.delete("key")
    await asyncio.sleep(0.05)  # skip init signal about invalidation
    assert await caches.get("key") is None

    await caches.close()
    await cachef.close()


async def test_rewrite_bcast(create_cache):
    cachef_local = Memory()
    cachef = await create_cache(cachef_local)
    caches_local = Memory()
    caches = await create_cache(caches_local)

    await cachef.set("key", b"value")
    await asyncio.sleep(0.05)  # skip init signal about invalidation

    assert await cachef.get("key") == b"value"
    assert await caches.get("key") == b"value"

    await caches.set("key", b"new", expire=0.1)
    await asyncio.sleep(0.05)  # skip init signal about invalidation

    assert await cachef.get("key") == b"new"

    await asyncio.sleep(0.15)
    assert await caches.get("key") is None
    assert await cachef.get("key") is None

    assert await cachef.incr("key") == 1
    assert await caches.incr("key") == 2

    await caches.close()
    await cachef.close()


async def test_simple_cmd_bcast_many(create_cache):
    from cashews.backends.redis.client_side import _empty_in_redis

    local = Memory()
    cache = await create_cache(local)
    await cache.set("key:1", "test")
    await cache.set("key:2", "test2")

    await cache.get("key:1")
    assert await local.get("key:1") == "test"

    assert await cache.get_many("key:1", "key:2", "key:3") == ("test", "test2", None)
    assert await local.get("key:2") == "test2"
    assert await local.get("key:3") is _empty_in_redis

    async for key in cache.scan("key:*"):
        assert key in ("key:1", "key:2")
        break
    else:
        raise AssertionError()

    async for key, value in cache.get_match("key:*"):
        assert key in ("key:1", "key:2")
        assert value in ("test", "test2")
        break
    else:
        raise AssertionError()

    await local.clear()

    assert await cache.get_many("key:1", "key:3") == ("test", None)

    async for key in cache.scan("key:*"):
        assert key in ("key:1", "key:2")

    async for key, value in cache.get_match("key:*"):
        assert key in ("key:1", "key:2")
        assert value in ("test", "test2")

    assert await local.get("key:1") == "test"

    await cache.delete_match("key:*")
    assert await cache.get("key:1") is None
    assert await local.get("key:1") is _empty_in_redis

    async for _ in cache.scan("key:*"):
        raise AssertionError()

    async for _ in cache.get_match("key:*"):
        raise AssertionError()

    await cache.close()


@pytest.mark.parametrize("client_side_listen_timeout", [None, 0.1])
async def test_unsafe_redis_down(client_side_listen_timeout: float | None):
    from cashews.backends.redis.client_side import BcastClientSide

    cache = BcastClientSide(
        client_side_listen_timeout=client_side_listen_timeout,
        suppress=False,
        address="redis://localhost:9223",
    )

    init_start = time.time()
    with pytest.raises(asyncio.TimeoutError):
        await cache.init()
    init_end = time.time()

    if client_side_listen_timeout is None:
        # defaults to socket_timeout, which defaults to 1.0
        client_side_listen_timeout = 1.0

    assert client_side_listen_timeout <= init_end - init_start <= client_side_listen_timeout * 5

    with contextlib.suppress(asyncio.CancelledError):
        await cache.close()


async def test_safe_redis_down():
    from cashews.backends.redis.client_side import BcastClientSide

    cache = BcastClientSide(address="redis://localhost:9223")

    await cache.init()

    with contextlib.suppress(asyncio.CancelledError):
        await cache.close()


async def test_set_get_mem_overload(create_cache):
    cache = await create_cache(Memory(size=1))

    assert await cache.set("key1", "test")
    assert await cache.set("key2", "test")
    assert await cache.set("key1", "test2")

    assert await cache.get("key1", "test2")
    assert await cache.set("key2", "test")

    await cache.close()


async def test_set_tag_get(create_cache):
    backend = await create_cache()
    cache = Cache()
    cache._add_backend(backend)

    cache.register_tag("tag", "key{i}")
    assert await cache.set("key1", "test", expire=0.1, tags=["tag"])
    assert await cache.set("key2", "test", expire=0.1, tags=["tag"])
    assert await cache.set_pop("_tag:tag", count=1)
    await asyncio.sleep(0.11)
    assert not await cache.set_pop("_tag:tag", count=1)

    await cache.close()



from __future__ import annotations

import asyncio
from random import random
from unittest.mock import AsyncMock

import pytest

from cashews import Cache, key_context


def test_register_tags(cache: Cache):
    cache.register_tag("tag", "key")
    cache.register_tag("tag_template", "key{i}")
    cache.register_tag("tag_test:{i}", "key{i:len}:test")
    cache.register_tag("tag_func:{i:hash}", "key{i}:test")
    cache.register_tag("tag_context:{val}", "key{i}:test")
    cache.register_tag("!@#$%^&*():{i}", "!@#$%^&*(){i}:test")

    with key_context(val=10):
        assert cache.get_key_tags("key1:test") == [
            "tag_template",
            "tag_test:1",
            "tag_func:c4ca4238a0b923820dcc509a6f75849b",
            "tag_context:10",
        ]
    assert cache.get_key_tags("keyanytest") == ["tag_template"]
    assert cache.get_key_tags("keytest") == ["tag_template"]
    assert cache.get_key_tags("key") == ["tag", "tag_template"]
    assert cache.get_key_tags("k") == []
    assert cache.get_key_tags("prefixkey") == []
    assert cache.get_key_tags(":key") == []
    assert cache.get_key_tags("!@#$%^&*()$:test") == ["!@#$%^&*():$"]


async def test_tags_no_delete(cache: Cache):
    cache.register_tag("3", "key3")
    await cache.set("key3", "value", tags=["3"])
    await cache.delete_tags("2")
    assert await cache.get("key3") == "value"


async def test_tags_set(cache: Cache):
    cache.register_tag("1", "key{i}")
    cache.register_tag("2", "key{i}")
    cache.register_tag("3", "key{i}")
    await cache.set("key1", "value", tags=["1", "2"])
    await cache.set("key2", "value", tags=["2"])
    await cache.set("key3", "value", tags=["3"])

    assert await cache.get_many("key1", "key2", "key3") == ("value", "value", "value")

    await cache.delete_tags("2")
    assert await cache.get_many("key1", "key2", "key3") == (None, None, "value")

    await cache.delete_tags("1")
    assert await cache.get_many("key1", "key2", "key3") == (None, None, "value")

    await cache.delete_tags("3")
    assert await cache.get_many("key1", "key2", "key3") == (None, None, None)


async def test_tags_incr(cache: Cache):
    cache.register_tag("tag", "key10")
    await cache.incr("key10", tags=["tag"])
    assert await cache.get("key10") == 1

    await cache.delete_tags("tag")
    assert await cache.get("key10") is None


async def test_set_delete_set_delete_tag(cache: Cache):
    cache.register_tag("tag", "key")
    await cache.set("key", "value", tags=["tag"])
    await cache.delete("key")  # should remove from tags
    await cache.set("key", "value2")  # without tag
    await cache.delete_tags("tag")

    assert await cache.get("key") == "value2"


async def test_set_expire_clean(cache: Cache):
    cache.register_tag("tag", "key{i}")
    assert await cache.set("key1", "test", expire=0.1, tags=["tag"])
    assert await cache.set("key2", "test", expire=0.1, tags=["tag"])
    assert await cache.set_pop("_tag:tag", count=1)
    await asyncio.sleep(0.11)
    assert not await cache.set_pop("_tag:tag")


@pytest.mark.xfail  # for now redis without client side and disk cache can not pass this test
async def test_set_expire_clean_2(cache: Cache):
    cache.register_tag("tag", "key{i}")
    assert await cache.set("key1", "test", expire=0.1, tags=["tag"])
    await asyncio.sleep(0.09)
    assert await cache.set("key2", "test", expire=0.1, tags=["tag"])
    await asyncio.sleep(0.02)
    # at this time first cache is expired and should not exist in set
    assert await cache.get("key1") is None
    assert await cache.get("key2") is not None
    assert list(await cache.set_pop("_tag:tag", count=1))[0] == "key2"
    await asyncio.sleep(0.1)
    assert not list(await cache.set_pop("_tag:tag"))


async def test_tags_gte_100(cache: Cache):
    cache.register_tag("all", "key_{i}")
    cache.register_tag("{i}", "key_{i}")

    for i in range(150):
        await cache.set(f"key_{i}", "value", tags=[f"{i}", "all"])

    await cache.delete_tags("1", "2", "100")
    await cache.delete_tags("1", "2", "100")

    await cache.delete_tags("5", "all")


async def test_tag_decorator(cache: Cache):
    @cache(ttl="1m", key="key:{a}:{b}", tags=["all", "tag:{a}"])
    async def func(a, b=None):
        return random()

    first = await func(1)
    second = await func(2)
    one_more = await func(5)

    await cache.delete_tags("tag:1")

    assert await func(1) != first
    assert await func(2) == second

    await cache.delete_tags("all")

    assert await func(2) != second
    assert await func(5) != one_more


async def test_tag_soft_decorator(cache: Cache):
    @cache.soft(ttl="2m", key="key:{a}", tags=["all", "tag:{a}"])
    async def func(a):
        return random()

    first = await func(1)
    second = await func(2)
    one_more = await func(5)

    await cache.delete_tags("tag:1")

    assert await func(1) != first
    assert await func(2) == second

    await cache.delete_tags("all")

    assert await func(2) != second
    assert await func(5) != one_more


async def test_tag_early_decorator(cache: Cache):
    @cache.early(ttl="2m", early_ttl="10s", key="key:{a}", tags=["all", "tag:{a}"])
    async def func(a):
        return random()

    first = await func(1)
    second = await func(2)
    one_more = await func(5)

    await cache.delete_tags("tag:1")

    assert await func(1) != first
    assert await func(2) == second

    await cache.delete_tags("all")

    assert await func(2) != second
    assert await func(5) != one_more


async def test_tag_hit_decorator(cache: Cache):
    if cache.name == "diskcache":
        pytest.skip("there is race condition on set_add/set_remove implementation")

    @cache.dynamic(ttl="2m", key="key:{a}", tags=["all", "tag:{a}"])
    async def func(a):
        return random()

    first = await func(1)
    second = await func(2)
    one_more = await func(5)

    await cache.delete_tags("tag:1")

    assert await func(1) != first
    assert await func(2) == second

    await cache.delete_tags("all")

    assert await func(2) != second
    assert await func(5) != one_more


async def test_double_decorator(cache: Cache):
    @cache(ttl="1m", key="key:{a}", tags=["all"])
    @cache.soft(ttl="1m", key="key:{a}", tags=["all", "tag:{a}"])
    async def func(a):
        return random()

    second = await func(2)
    one_more = await func(5)

    await cache.delete_tags("all")

    assert await func(2) != second
    assert await func(5) != one_more


@pytest.mark.redis
async def test_delete_tags_separate_backend(cache: Cache, redis_dsn: str):
    tag_backend = cache.setup_tags_backend(redis_dsn)
    tag_backend.set_pop = AsyncMock(side_effect=[["key", "key2"], []])

    cache.register_tag(tag="tag", key_template="key")
    await cache.init()

    await cache.delete_tags("tag")

    tag_backend.set_pop.assert_awaited_with(key="_tag:tag", count=100)

    await tag_backend.close()


async def test_templated_tag_with_none_value(cache: Cache):
    @cache(
        ttl=None,
        tags=["a:{a};b:{b}"],
    )
    async def cached(
        a: int,
        b: int | None = None,
    ) -> str:
        return f"{a}{b}"

    assert await cached(1) == "1None"


async def test_template_tag_function(cache: Cache):
    @cache(ttl="1m", key="{user.name:hash}:func", tags=["tag:{user.name}", "tag:{user.name:hash}"])
    async def func(user):
        return random()

    class User:
        def __init__(self, name):
            self.name = name

        def __hash__(self):
            return hash(self.name)

    await func(User("test"))
    await cache.delete_tags("tag:test")



from __future__ import annotations

import warnings
from collections.abc import Iterator
from contextlib import contextmanager
from contextvars import ContextVar
from typing import Any

_REWRITE = "__rewrite"
TContext = dict[str, Any]
_template_context: ContextVar[TContext | None] = ContextVar("template_context", default=None)
_empty = object()


@contextmanager
def context(rewrite=_empty, **values) -> Iterator[None]:
    if rewrite is not _empty:
        warnings.warn(
            "`rewrite` deprecated and will be removed in next release, use @ notation",
            DeprecationWarning,
            stacklevel=2,
        )
    else:
        rewrite = False
    new_context = {**_get_raw(), **values, _REWRITE: rewrite}
    token = _template_context.set(new_context)
    try:
        yield
    finally:
        _template_context.reset(token)


def get() -> tuple[TContext, bool]:
    _context = {**_get_raw()}  # a copy
    return _context, _context.pop(_REWRITE)


def _get_raw() -> TContext:
    return _template_context.get() or {_REWRITE: False}


def register(*names: str) -> None:
    warnings.warn(
        "`register_key_context` deprecated and will be removed in next release, use @ notation",
        DeprecationWarning,
        stacklevel=2,
    )
    new_names = dict.fromkeys(names, "")
    _template_context.set({**new_names, **_get_raw()})



from __future__ import annotations

from contextlib import AbstractContextManager
from typing import ClassVar

from cashews.backends.interface import _BackendInterface
from cashews.decorators import CacheDetect, context_cache_detect

from .backend_settings import register_backend  # noqa
from .callback import CallbackWrapper
from .decorators import DecoratorsWrapper
from .disable_control import ControlWrapper
from .tags import CommandsTagsWrapper
from .transaction import TransactionMode, TransactionWrapper

__all__ = [
    "Cache",
    "TransactionMode",
    "register_backend",
]


class Cache(
    TransactionWrapper,
    ControlWrapper,
    CallbackWrapper,
    CommandsTagsWrapper,
    DecoratorsWrapper,
    _BackendInterface,
):
    detect: ClassVar[AbstractContextManager[CacheDetect]] = context_cache_detect



import asyncio

from cashews._typing import AsyncCallable_T, Middleware, Result_T
from cashews.backends.interface import Backend
from cashews.commands import Command


def create_auto_init() -> Middleware:
    lock = asyncio.Lock()

    async def _auto_init(call: AsyncCallable_T, cmd: Command, backend: Backend, *args, **kwargs) -> Result_T:
        if backend.is_init:
            return await call(*args, **kwargs)
        async with lock:
            if not backend.is_init:
                await backend.init()

        return await call(*args, **kwargs)

    return _auto_init



from __future__ import annotations

import contextlib
import uuid
from collections.abc import Iterator
from typing import TYPE_CHECKING, Any

from cashews.commands import PATTERN_CMDS, Command

from .wrapper import Wrapper

if TYPE_CHECKING:  # pragma: no cover
    from cashews._typing import AsyncCallable_T, Callback, Key, Result_T, ShortCallback
    from cashews.backends.interface import Backend


class CallbackMiddleware:
    def __init__(self) -> None:
        self._callbacks: dict[str, Callback] = {}

    async def __call__(self, call: AsyncCallable_T, cmd: Command, backend: Backend, *args, **kwargs) -> Result_T:
        result = await call(*args, **kwargs)
        if not self._callbacks:
            return result
        as_key = "pattern" if cmd in PATTERN_CMDS else "key"
        key = kwargs.get(as_key)
        if key is None or result is None:
            return result
        for callback in self._callbacks.values():
            await callback(cmd, key=key, result=result, backend=backend)
        return result

    def add_callback(self, callback: Callback, name: str) -> None:
        self._callbacks[name] = callback

    def remove_callback(self, name: str) -> None:
        del self._callbacks[name]

    @contextlib.contextmanager
    def callback(self, callback: Callback) -> Iterator[None]:
        name = uuid.uuid4().hex
        self.add_callback(callback, name)
        try:
            yield
        finally:
            self.remove_callback(name)


class CallbackWrapper(Wrapper):
    def __init__(self, name: str = ""):
        super().__init__(name)
        self.callbacks = CallbackMiddleware()
        self.add_middleware(self.callbacks)

    @contextlib.contextmanager
    def callback(self, callback: ShortCallback, cmd: Command) -> Iterator[None]:
        t_cmd = cmd

        async def _wrapped_callback(cmd: Command, key: Key, result: Any, backend: Backend) -> None:
            if cmd == t_cmd:
                callback(key, result=result)

        with self.callbacks.callback(_wrapped_callback):
            yield



from __future__ import annotations

from collections.abc import Awaitable, Iterable
from datetime import timedelta
from typing import TYPE_CHECKING, Any, Callable, Protocol, TypeVar, Union

if TYPE_CHECKING:  # pragma: no cover
    from . import Command
    from .backends.interface import Backend


_TTLTypes = Union[int, float, str, timedelta, None]
TTL = Union[_TTLTypes, Callable[..., _TTLTypes]]


class CallableCacheCondition(Protocol):
    def __call__(
        self, result: Any, args: tuple, kwargs: dict[str, Any], key: str = ""
    ) -> bool | Exception:  # pragma: no cover
        ...


Key = str
KeyTemplate = str
KeyOrTemplate = Union[KeyTemplate, Key]
Value = Any
Default = TypeVar("Default")
Tag = str
Tags = Iterable[Tag]
Exceptions = Union[type[Exception], Iterable[type[Exception]], None]

CacheCondition = Union[CallableCacheCondition, str, None]

Result_T = TypeVar("Result_T")
AsyncCallable_T = Callable[..., Awaitable[Result_T]]
Callable_T = Callable[..., Result_T]

DecoratedFunc = TypeVar("DecoratedFunc", bound=AsyncCallable_T)


class Middleware(Protocol):
    def __call__(
        self,
        call: AsyncCallable_T,
        cmd: Command,
        backend: Backend,
        *args,
        **kwargs,
    ) -> Awaitable[Result_T | None]:  # pragma: no cover
        ...


class OnRemoveCallback(Protocol):
    async def __call__(
        self,
        keys: Iterable[Key],
        backend: Backend,
    ) -> None:  # pragma: no cover
        ...


class Callback(Protocol):
    async def __call__(self, cmd: Command, key: Key, result: Any, backend: Backend) -> None:
        pass


class ShortCallback(Protocol):
    def __call__(self, key: Key, result: Any) -> None:
        pass


class ICustomEncoder(Protocol):
    async def __call__(
        self, value: Value, backend: Backend, key: Key, expire: float | None
    ) -> bytes:  # pragma: no cover
        ...


class ICustomDecoder(Protocol):
    async def __call__(self, value: bytes, backend: Backend, key: Key) -> Value:  # pragma: no cover
        ...



from __future__ import annotations

import zlib
from collections.abc import MutableSequence

algorithms: MutableSequence = [
    zlib.crc32,
]

try:
    import xxhash
except ImportError:
    pass
else:
    algorithms.extend(
        [
            xxhash.xxh3_64_intdigest,
            xxhash.xxh64_intdigest,
            xxhash.xxh3_128_intdigest,
            xxhash.xxh32_intdigest,
        ]
    )


def get_indexes(key: str, number_of_buckets: int, max_index: int) -> set[int]:
    """
    return array with bit indexes for given value (key) [23, 45, 15]
    """

    assert max_index >= number_of_buckets

    indexes = set()
    for i in range(number_of_buckets):
        ii = i % len(algorithms)

        value = algorithms[ii](f"{key}_{i}".encode()) % max_index
        while value in indexes:
            i += 1
            value = algorithms[ii](f"{key}_{i}".encode()) % max_index
        indexes.add(value)
    return indexes



from __future__ import annotations

import time
from functools import wraps
from typing import TYPE_CHECKING, Callable

from cashews.backends.interface import _BackendInterface
from cashews.key import get_cache_key, get_cache_key_template
from cashews.ttl import ttl_to_seconds

from ._exception import RaiseException, return_or_raise
from .defaults import context_cache_detect

if TYPE_CHECKING:  # pragma: no cover
    from cashews._typing import TTL, CallableCacheCondition, DecoratedFunc

__all__ = ("iterator",)


if "anext" not in globals():

    async def anext(ait):
        return await ait.__anext__()


def iterator(
    backend: _BackendInterface,
    ttl: TTL,
    key: str | None = None,
    condition: CallableCacheCondition = lambda *args, **kwargs: True,
) -> Callable[[DecoratedFunc], DecoratedFunc]:
    """
    Cache decorator for iterators
    execute wrapped call and store a result with ttl if condition return true
    :param backend: cache backend
    :param ttl: duration in seconds to store a result or a callable
    :param key: custom cache key, may contain alias to args or kwargs passed to a call
    :param condition: callable object that determines whether the result will be saved or not
    """

    ttl = ttl_to_seconds(ttl)

    def _decor(async_iterator: DecoratedFunc) -> DecoratedFunc:
        _key_template = get_cache_key_template(async_iterator, key=key)

        @wraps(async_iterator)
        async def _wrap(*args, **kwargs):
            _ttl = ttl_to_seconds(ttl, *args, **kwargs, with_callable=True)
            _cache_key = get_cache_key(async_iterator, _key_template, args, kwargs)

            cached = await backend.get(_cache_key)
            chunk_number = 0
            if cached:
                context_cache_detect._set(
                    _cache_key,
                    ttl=_ttl,
                    name="iterator",
                    template=_key_template,
                    value=cached,
                )
                while True:
                    chunk = await backend.get(_cache_key + f":{chunk_number}")
                    if not chunk:
                        return
                    yield return_or_raise(chunk)
                    chunk_number += 1

            start = time.monotonic()
            _to_cache = False
            _async_iterator = async_iterator(*args, **kwargs)
            while True:
                try:
                    chunk = await anext(_async_iterator)
                except StopAsyncIteration:
                    break
                except Exception as exc:
                    cond_res = condition(exc, args, kwargs, key=_cache_key)
                    if cond_res and isinstance(cond_res, Exception):
                        _to_cache = True
                        await backend.set(_cache_key + f":{chunk_number}", RaiseException(exc), expire=_ttl)
                        await backend.set(_cache_key, True, expire=_ttl - time.monotonic() + start)
                    raise exc
                yield chunk
                if condition(chunk, args, kwargs, key=_cache_key):
                    _to_cache = True
                    await backend.set(_cache_key + f":{chunk_number}", chunk, expire=_ttl)
                chunk_number += 1
            if _to_cache:
                executing_time = time.monotonic() - start
                await backend.set(_cache_key, True, expire=_ttl - executing_time)
            return

        return _wrap  # type: ignore[return-value]

    return _decor



from __future__ import annotations

from functools import partial
from typing import TYPE_CHECKING, Any, Callable

from cashews import validation
from cashews.backends.interface import Backend
from cashews.commands import Command
from cashews.exceptions import NotConfiguredError
from cashews.picklers import PicklerType
from cashews.serialize import get_serializer

from .auto_init import create_auto_init
from .backend_settings import settings_url_parse

if TYPE_CHECKING:  # pragma: no cover
    from cashews._typing import Key, Middleware


class Wrapper:
    default_prefix = ""

    def __init__(self, name: str = ""):
        self._backends: dict[str, Backend] = {}
        self._middlewares: dict[str, tuple[Middleware, ...]] = {}
        self._sorted_prefixes: tuple[str, ...] = ()
        self._default_middlewares: list[Middleware] = [
            create_auto_init(),
            validation._invalidate_middleware,
        ]
        self.name = name
        super().__init__()

    def add_middleware(self, middleware: Middleware) -> None:
        self._default_middlewares.append(middleware)

    def _get_backend(self, key: Key) -> Backend:
        for prefix in self._sorted_prefixes:
            if key.startswith(prefix):
                return self._backends[prefix]
        self._check_setup()
        raise NotConfiguredError("Backend for given key not configured")

    def _with_middlewares(self, cmd: Command, key: Key) -> Callable[..., Any]:
        backend = self._get_backend(key)
        middlewares = [*self._default_middlewares, *self._middlewares[backend._id]]
        return self._with_middlewares_for_backend(cmd, backend, middlewares)

    def _with_middlewares_for_backend(
        self, cmd: Command, backend: Backend, middlewares: list[Middleware]
    ) -> Callable[..., Any]:
        call = getattr(backend, cmd.value)
        for middleware in middlewares:
            call = partial(middleware, call, cmd, backend)
        return call

    def setup(
        self,
        settings_url: str,
        middlewares: tuple[Middleware, ...] = (),
        prefix: str = default_prefix,
        **kwargs: Any,
    ) -> Backend:
        backend_class, params, pickle_type = settings_url_parse(settings_url)
        params.update(kwargs)

        disable = params.pop("disable") if "disable" in params else not params.pop("enable", True)

        serializer = get_serializer(
            secret=params.pop("secret", None),
            digestmod=params.pop("digestmod", b"md5"),
            check_repr=params.pop("check_repr", True),
            pickle_type=PicklerType(params.pop("pickle_type", pickle_type)),
        )
        backend = backend_class(**params, serializer=serializer)
        if disable:
            backend.disable()
        self._add_backend(backend, middlewares, prefix)
        return backend

    def is_setup(self) -> bool:
        return bool(self._backends)

    def _check_setup(self) -> None:
        if not self._backends:
            raise NotConfiguredError("run `cache.setup(...)` before using cache")

    def _add_backend(
        self, backend: Backend, middlewares: tuple[Middleware, ...] = (), prefix: str = default_prefix
    ) -> None:
        self._backends[prefix] = backend
        self._middlewares[backend._id] = middlewares
        self._sorted_prefixes = tuple(sorted(self._backends.keys(), reverse=True))

    async def init(self, *args: Any, **kwargs: Any) -> None:
        if args or kwargs:
            self.setup(*args, **kwargs)
        for backend in self._backends.values():
            await backend.init()

    @property
    def is_init(self) -> bool:
        return all(backend.is_init for backend in self._backends.values())

    async def close(self) -> None:
        for backend in self._backends.values():
            await backend.close()



from __future__ import annotations

import inspect
from collections.abc import Container, Iterable
from functools import lru_cache
from typing import TYPE_CHECKING, Any, Callable

from .exceptions import WrongKeyError
from .formatter import default_format, default_formatter

if TYPE_CHECKING:  # pragma: no cover
    from ._typing import Key, KeyOrTemplate, KeyTemplate

_KWARGS = "__kwargs__"
_ARGS = "__args__"
_ARGS_KWARGS = (_ARGS, _KWARGS)
Args = tuple[Any, ...]
Kwargs = dict[str, Any]


def get_cache_key(
    func: Callable,
    template: KeyTemplate | None = None,
    args: Args = (),
    kwargs: Kwargs | None = None,
) -> Key:
    """
    Get cache key name for function (:param func) called with args and kwargs
    if func_args is passed key build with parameters are included in func_args dict or tuple otherwise use all of them
    Used function module and name as prefix if key parameter not passed
    :param func: Target function
    :param template: precompile template
    :param args: call positional arguments
    :param kwargs: call keyword arguments
    :return: cache key for call
    """
    kwargs = kwargs or {}
    if not args and template and _KWARGS not in template and _ARGS not in template:
        key_values = kwargs
    else:
        key_values = _get_call_values(func, args, kwargs)
    _key_template = template or get_cache_key_template(func)
    return default_format(_key_template, **key_values)


def get_func_params(func: Callable) -> Iterable[str]:
    signature = _get_func_signature(func)
    for param_name, param in signature.parameters.items():
        if param.kind == inspect.Parameter.VAR_KEYWORD:
            yield _KWARGS
        elif param.kind == inspect.Parameter.VAR_POSITIONAL:
            yield _ARGS
        else:
            yield param_name


@lru_cache(maxsize=10000)
def get_cache_key_template(
    func: Callable,
    key: KeyOrTemplate | None = None,
    prefix: str = "",
    exclude_parameters: Container = (),
) -> KeyOrTemplate:
    """
    Get cache key name for function (:param func) called with args and kwargs
    Used function module and name as prefix if key parameter not passed
    :param func: Target function
    :param key: template for key, may contain alias to args or kwargs passed to a call
    :param prefix: a prefix
    :param exclude_parameters: array of `args` and `kwargs` names to exclude from a key
        template (if key parameter not passed)
    :return: cache key template
    """

    if key is None:
        key = generate_key_template(func, exclude_parameters)
    else:
        if "{" in key:
            _check_key_params(key, get_func_params(func))
    if prefix:
        key = f"{prefix}:{key}"
    return key


def generate_key_template(func: Callable, exclude_parameters: Container = ()) -> KeyTemplate:
    """
    Generate template for function (:param func) called with args and kwargs
    Used function module and name as prefix if key parameter not passed
    :param func: Target function
    :param exclude_parameters: array of `args` and `kwargs` names to exclude
        from a key template (if key parameter not passed)
    :return: cache key template
    """
    func_params = tuple(get_func_params(func))
    key_template = f"{func.__module__}:{func.__name__}"
    if func_params and func_params[0] == "self":
        key_template = f"{func.__module__}:{func.__qualname__}"
    for param_name in func_params:
        if param_name in exclude_parameters:
            continue
        if param_name in _ARGS_KWARGS:
            key_template += f":{{{param_name}}}"
        else:
            key_template += f":{param_name}:{{{param_name}}}"
    return key_template


class _Star:
    def __getattr__(self, item):
        return self

    def __getitem__(self, item):
        return self

    def __call__(self, *args, **kwargs):
        return "*"


def _check_key_params(key: KeyOrTemplate, func_params: Iterable[str]):
    func_params = {param: _Star() for param in func_params}

    def _default(name):
        raise WrongKeyError(f"Wrong parameter placeholder '{name}' in the key ")

    with default_formatter.default(_default):
        default_format(key, **func_params)


def get_call_values(func: Callable, args: Args, kwargs: Kwargs) -> dict:
    """
    Return dict with arguments and their values for function call with given positional and keywords arguments
    :param func: Target function
    :param args: call positional arguments
    :param kwargs: call keyword arguments
    """
    key_values = {}
    for _key, _value in _get_call_values(func, args, kwargs).items():
        if _key not in _ARGS_KWARGS:
            key_values[_key] = _value
    return key_values


@lru_cache(maxsize=1000)
def _get_func_signature(func: Callable):
    return inspect.signature(func)


def _get_call_values(func: Callable, args: Args, kwargs: Kwargs):
    if not args:
        _kwargs = {**kwargs}
        for name, parameter in _get_func_signature(func).parameters.items():
            if parameter.kind != inspect.Parameter.VAR_KEYWORD and name in _kwargs:
                del _kwargs[name]
        return {**kwargs, _KWARGS: _kwargs}

    signature = _get_func_signature(func).bind(*args, **kwargs)
    signature.apply_defaults()
    result = {}
    for _name, _value in signature.arguments.items():
        parameter: inspect.Parameter = signature.signature.parameters[_name]  # type: ignore[no-redef]
        if parameter.kind == inspect.Parameter.VAR_KEYWORD:
            result[_KWARGS] = _value
            result.update(_value)
        elif parameter.kind == inspect.Parameter.VAR_POSITIONAL:
            result[_ARGS] = _value
        else:
            result[_name] = _value
    return result


def noself(decor_func):
    def _decor(*args, **kwargs):
        def outer(method):
            if "key" not in kwargs:
                kwargs["key"] = get_cache_key_template(method, exclude_parameters=("self",))
            return decor_func(*args, **kwargs)(method)

        return outer

    return _decor



from __future__ import annotations

import hashlib
import hmac
from typing import TYPE_CHECKING

from .exceptions import SignIsMissingError, UnSecureDataError
from .picklers import Pickler, PicklerType, get_pickler

if TYPE_CHECKING:  # pragma: no cover
    from ._typing import ICustomDecoder, ICustomEncoder, Key, Value
    from .backends.interface import Backend


def _seal(digestmod):
    def sign(key: bytes, value: bytes) -> bytes:
        return hmac.new(key, value, digestmod).hexdigest().encode()

    return sign


def simple_sign(key: bytes, value: bytes) -> bytes:
    s = sum(key) + sum(value)
    return f"{s:x}".encode()


def _to_bytes(value: str | bytes) -> bytes:
    if isinstance(value, str):
        value = value.encode()
    return value


class HashSigner:
    _digestmods = {
        b"sha1": _seal(hashlib.sha1),
        b"md5": _seal(hashlib.md5),
        b"sha256": _seal(hashlib.sha256),
        b"sum": simple_sign,
    }

    def __init__(self, secret: str | bytes, digestmod: str | bytes = b"md5"):
        self._secret = _to_bytes(secret)
        self._digestmod = _to_bytes(digestmod)

    def sign(self, key: Key, value: bytes) -> bytes:
        sign = self._gen_sign(key, value, self._digestmod)
        return self._digestmod + b":" + sign + b"_" + value

    def check_sign(self, key: Key, value: bytes) -> bytes:
        try:
            sign, value = value.split(b"_", maxsplit=1)
        except ValueError as exc:
            raise SignIsMissingError(f"key: {key}") from exc

        sign, digestmod = self._get_sign_and_digestmod(sign)
        expected_sign = self._gen_sign(key, value, digestmod)
        if expected_sign != sign:
            raise UnSecureDataError(f"{expected_sign!r} != {sign!r}")
        return value

    def _gen_sign(self, key: Key, value: bytes, digestmod: bytes) -> bytes:
        value = key.encode() + value
        return self._digestmods[digestmod](self._secret, value)

    def _get_sign_and_digestmod(self, sign: bytes) -> tuple[bytes, bytes]:
        digestmod = self._digestmod
        if b":" in sign:
            digestmod, sign = sign.split(b":", maxsplit=1)
        if digestmod not in self._digestmods:
            raise UnSecureDataError()
        return sign, digestmod


class NullSigner:
    @staticmethod
    def sign(key: Key, value: bytes) -> bytes:
        return value

    @staticmethod
    def check_sign(key: Key, value: bytes) -> bytes:
        return value


class Serializer:
    _type_mapping: dict[bytes, tuple[ICustomEncoder, ICustomDecoder]] = {}

    def __init__(self, check_repr=False):
        self._check_repr = check_repr
        self._pickler = get_pickler(PicklerType.NULL)
        self._signer = NullSigner()

    def set_signer(self, signer):
        self._signer = signer

    def set_pickler(self, pickler):
        self._pickler = pickler

    @classmethod
    def register_type(cls, klass: type, encoder, decoder):
        cls._type_mapping[bytes(klass.__name__, "utf8")] = (encoder, decoder)

    async def encode(self, backend: Backend, key: Key, value: Value, expire: float | None) -> bytes:  # on SET
        if isinstance(value, int) and not isinstance(value, bool):
            return value  # type: ignore[return-value]
        _value = await self._custom_encode(backend, key, value, expire)
        if _value is not None:
            return self._signer.sign(key, _value)
        return self._signer.sign(key, self._pickler.dumps(value))

    async def _custom_encode(self, backend, key: Key, value: Value, expire: float | None) -> bytes | None:
        value_type = bytes(type(value).__name__, "utf8")
        if value_type not in self._type_mapping:
            return None
        encoder, _ = self._type_mapping[value_type]
        encoded_value = await encoder(value, backend, key, expire)
        return value_type + b":" + encoded_value

    async def decode(self, backend: Backend, key: Key, value: bytes, default: Value) -> Value:  # on GET
        if value is default:
            return default
        if not isinstance(value, bytes):
            return value
        if value.isdigit():
            return int(value)
        try:
            value = self._signer.check_sign(key, value)
        except SignIsMissingError:
            return default

        try:
            value = self._decode(value)
        except self._pickler.UnpicklingError:
            pass
        except AttributeError:
            return default
        if isinstance(value, bytes):
            return await self._custom_decode(backend, key, value, default)
        return value

    def _decode(self, value: bytes) -> Value:
        value = self._pickler.loads(value)
        if self._check_repr:
            repr(value)
        return value

    async def _custom_decode(self, backend: Backend, key: Key, value: bytes, default: Value) -> Value:
        try:
            value_type, value = value.split(b":", maxsplit=1)
        except ValueError:
            return default
        if value_type not in self._type_mapping:
            return default
        _, decoder = self._type_mapping[value_type]
        try:
            return await decoder(value, backend, key)
        except DecodeError:
            return default


class DecodeError(Exception):
    pass


register_type = Serializer.register_type


async def bytes_encoder(value: bytes, *args, **kwargs):
    return value


async def bytes_decoder(value: bytes, *args, **kwargs):
    return value


register_type(bytes, bytes_encoder, bytes_decoder)


def get_serializer(
    secret: str | bytes | None = None,
    digestmod: str | bytes = b"md5",
    check_repr: bool = True,
    pickle_type: PicklerType | None = None,
) -> Serializer:
    _serializer = Serializer(check_repr=check_repr)
    if secret:
        _serializer.set_signer(HashSigner(secret, digestmod))
    _serializer.set_pickler(_get_pickler(pickle_type or PicklerType.NULL, bool(secret)))
    return _serializer


def _get_pickler(pickle_type: PicklerType, hash_key: bool) -> Pickler:
    if pickle_type is PicklerType.NULL and hash_key:
        pickle_type = PicklerType.DEFAULT
    return get_pickler(pickle_type)


DEFAULT_SERIALIZER = get_serializer(pickle_type=PicklerType.DEFAULT)
NULL_SERIALIZER = get_serializer(pickle_type=PicklerType.NULL)



from __future__ import annotations

import asyncio
import time
from collections.abc import AsyncIterator, Iterable, Mapping
from typing import Any
from uuid import uuid4

from cashews import LockedError
from cashews._typing import Key, OnRemoveCallback, Value
from cashews.backends.interface import NOT_EXIST, UNLIMITED, Backend
from cashews.backends.memory import Memory

_empty = object()
_GLOBAL_LOCK_KEY = ":serializable:lock"
_LOCK_PREFIX = ":tx_lock"


class TransactionBackend(Backend):
    __slots__ = [
        "_backend",
        "_local_cache",
        "_to_delete",
        "__disable",
        "_id",
    ]

    def __init__(self, backend: Backend):
        self._backend = backend
        self._local_cache = Memory()
        self._to_delete: set[Key] = set()
        super().__init__()
        self._id = backend._id

    def _key_is_delete(self, key: Key) -> bool:
        return key in self._to_delete

    async def commit(self):
        if self._to_delete:
            await self._backend.delete_many(*self._to_delete)

        expire_group = {}
        for key, (expire, value) in self._local_cache.store.items():
            if expire:
                expire = int(expire - time.time())
            expire_group.setdefault(expire, {})[key] = value

        for expire, kv in expire_group.items():
            await self._backend.set_many(kv, expire=expire)
        self._clear_local_storage()

    async def rollback(self):
        self._clear_local_storage()

    def _clear_local_storage(self):
        self._local_cache = Memory()
        self._to_delete = set()

    def on_remove_callback(self, callback: OnRemoveCallback):
        self._backend.on_remove_callback(callback)
        self._local_cache.on_remove_callback(callback)

    async def set(
        self,
        key: Key,
        value: Value,
        expire: float | None = None,
        exist: bool | None = None,
    ) -> bool:
        if (
            exist is not None
            and await self._backend.exists(key) is not exist
            and await self._local_cache.exists(key) is not exist
        ):
            return False
        self._to_delete.discard(key)
        return await self._local_cache.set(key, value, expire, exist)

    async def set_many(self, pairs: Mapping[Key, Value], expire: float | None = None):
        self._to_delete.difference_update(pairs.keys())
        return await self._local_cache.set_many(pairs, expire)

    async def incr(self, key: Key, value: int = 1, expire: float | None = None) -> int:
        if not await self._local_cache.exists(key) and key not in self._to_delete:
            current = await self._backend.get(key, 0)
            await self._local_cache.set(key, current)
        self._to_delete.discard(key)
        return await self._local_cache.incr(key, value, expire=expire)

    async def delete(self, key: Key) -> bool:
        await self._local_cache.delete(key)
        self._to_delete.add(key)
        return True

    async def delete_many(self, *keys: Key):
        await self._local_cache.delete_many(*keys)
        self._to_delete.update(keys)

    async def delete_match(self, pattern: str):
        await self._local_cache.delete_match(pattern)
        async for key in self._backend.scan(pattern):
            self._to_delete.add(key)

    async def expire(self, key: Key, timeout: float):
        if self._key_is_delete(key):
            return
        value = await self._backend.get(key, default=_empty)
        if value is _empty:
            return await self._local_cache.expire(key, timeout)
        await self._local_cache.set(key, value, expire=timeout)

    # non transaction - proxy methods with custom logic
    async def get(self, key: str, default: Value | None = None) -> Value:
        if self._key_is_delete(key):
            return default
        value = await self._local_cache.get(key, default=_empty)
        if value is not _empty:
            return value
        return await self._backend.get(key, default=default)

    async def get_many(self, *keys: Key, default: Value | None = None) -> tuple[Value | None, ...]:
        missed_keys = set(keys)
        values = dict.fromkeys(keys, default)

        _keys = list(missed_keys)
        for i, value in enumerate(await self._local_cache.get_many(*_keys, default=_empty)):
            if value is not _empty:
                key = _keys[i]
                values[key] = value
                missed_keys.remove(key)

        _keys = list(missed_keys)
        for i, value in enumerate(await self._backend.get_many(*_keys, default=default)):
            key = _keys[i]
            if not self._key_is_delete(key):
                values[key] = value
        return tuple(values[key] for key in keys)

    async def get_match(self, pattern: str, batch_size: int = 100) -> AsyncIterator[tuple[Key, Value]]:
        _local_state = set()
        async for key, value in self._local_cache.get_match(pattern):
            yield key, value
            _local_state.add(key)
        async for key, value in self._backend.get_match(pattern, batch_size=batch_size):
            if self._key_is_delete(key):
                continue
            if key in _local_state:
                continue
            yield key, value

    async def get_expire(self, key: Key) -> int:
        if self._key_is_delete(key):
            return NOT_EXIST
        local_expire = await self._local_cache.get_expire(key)
        if local_expire >= 0:
            return local_expire
        backend_expire = await self._backend.get_expire(key)
        if backend_expire is NOT_EXIST and local_expire is UNLIMITED:
            return UNLIMITED
        return backend_expire

    async def scan(self, pattern: str, batch_size: int = 100) -> AsyncIterator[Key]:
        _local_state = set()
        async for key in self._local_cache.scan(pattern):
            yield key
            _local_state.add(key)
        async for key in self._backend.scan(pattern, batch_size=batch_size):
            if self._key_is_delete(key):
                continue
            if key in _local_state:
                continue
            yield key

    async def exists(self, key: Key) -> bool:
        if await self._local_cache.exists(key):
            return True
        if self._key_is_delete(key):
            return False
        return await self._backend.exists(key)

    # non transaction - proxy methods
    async def get_keys_count(self) -> int:
        return await self._backend.get_keys_count() + await self._local_cache.get_keys_count()

    @property
    def is_init(self) -> bool:
        return self._backend.is_init

    async def init(self):
        return await self._backend.init()

    async def close(self):
        return await self._backend.close()

    async def set_raw(self, key: Key, value: Value, **kwargs: Any):
        return await self._backend.set_raw(key, value, **kwargs)

    async def get_raw(self, key: Key) -> Value:
        return await self._backend.get_raw(key)

    async def get_bits(self, key: Key, *indexes: int, size: int = 1) -> tuple[int, ...]:
        return await self._backend.get_bits(key, *indexes, size=size)

    async def incr_bits(self, key: Key, *indexes: int, size: int = 1, by: int = 1) -> tuple[int, ...]:
        return await self._backend.incr_bits(key, *indexes, size=size, by=by)

    async def slice_incr(
        self,
        key: Key,
        start: int | float,
        end: int | float,
        maxvalue: int,
        expire: float | None = None,
    ) -> int:
        return await self._backend.slice_incr(key, start, end, maxvalue, expire)

    async def get_size(self, key: Key) -> int:
        return await self._backend.get_size(key)

    async def ping(self, message: bytes | None = None) -> bytes:
        return await self._backend.ping(message)

    async def clear(self):
        self._to_delete = set()
        await self._local_cache.clear()
        return await self._backend.clear()

    async def set_lock(self, key: Key, value: Value, expire: float) -> bool:
        return await self._backend.set_lock(key, value, expire)

    async def is_locked(
        self,
        key: Key,
        wait: float | None = None,
        step: float = 0.1,
    ) -> bool:
        return await self._backend.is_locked(key, wait, step)

    async def unlock(self, key: Key, value: Value) -> bool:
        return await self._backend.unlock(key, value)

    async def set_add(self, key: Key, *values: str, expire: float | None = None):
        return await self._backend.set_add(key, *values, expire=expire)

    async def set_remove(self, key: Key, *values: str):
        return await self._backend.set_remove(key, *values)

    async def set_pop(self, key: Key, count: int = 100) -> Iterable[str]:
        return await self._backend.set_pop(key, count)


class LockTransactionBackend(TransactionBackend):
    __slots__ = [
        "_backend",
        "_local_cache",
        "_to_delete",
        "__disable",
        "_locks",
        "_lock_id",
        "_serializable",
        "_timeout",
    ]

    def __init__(self, backend: Backend, serializable=False, timeout=10):
        super().__init__(backend)
        self._locks: set[Key] = set()
        self._lock_id = uuid4().hex
        self._serializable = serializable
        self._timeout = timeout

    def _get_lock_key(self, key: Key) -> Key:
        if self._serializable:
            return _GLOBAL_LOCK_KEY
        return f"{_LOCK_PREFIX}:{key}"

    async def _lock_updates(self, key: Key):
        lock_key = self._get_lock_key(key)
        if lock_key in self._locks:
            return
        wait = self._timeout
        step = 0.1
        while wait > 0.0:
            wait -= step
            wait = round(wait, 1)
            if await self._backend.set_lock(lock_key, self._lock_id, expire=self._timeout):
                self._locks.add(lock_key)
                return
            if lock_key in self._locks:
                return
            await asyncio.sleep(step)
        raise LockedError("probably deadlock or long running transactions")

    async def _unlock_updates(self):
        locks = self._locks
        self._locks = set()
        if locks:
            await asyncio.gather(*[self._backend.unlock(key, self._lock_id) for key in locks])

    async def commit(self):
        try:
            await super().commit()
        finally:
            await self._unlock_updates()

    async def rollback(self):
        try:
            await super().rollback()
        finally:
            await self._unlock_updates()

    async def set(
        self,
        key: Key,
        value: Value,
        expire: float | None = None,
        exist: bool | None = None,
    ) -> bool:
        await self._lock_updates(key)
        return await super().set(key, value, expire=expire, exist=exist)

    async def set_many(self, pairs: Mapping[Key, Value], expire: float | None = None):
        for key in pairs:
            await self._lock_updates(key)
        res = await super().set_many(pairs, expire=expire)
        return res

    async def incr(self, key: Key, value: int = 1, expire: float | None = None) -> int:
        await self._lock_updates(key)
        return await super().incr(key, value, expire=expire)

    async def delete(self, key: Key) -> bool:
        await self._lock_updates(key)
        res = await super().delete(key)
        return res

    async def delete_many(self, *keys: Key):
        for key in keys:
            await self._lock_updates(key)
        res = await super().delete_many(*keys)
        return res

    async def delete_match(self, pattern: str):
        await self._local_cache.delete_match(pattern)
        async for key in self._backend.scan(pattern):
            await self._lock_updates(key)
            self._to_delete.add(key)

    async def expire(self, key: Key, timeout: float):
        await self._lock_updates(key)
        return await super().expire(key, timeout)



from __future__ import annotations

import random
from datetime import datetime, timezone
from functools import wraps
from typing import TYPE_CHECKING, Callable

from cashews.backends.interface import _BackendInterface
from cashews.exceptions import CircuitBreakerOpen
from cashews.key import get_cache_key, get_cache_key_template
from cashews.ttl import ttl_to_seconds

if TYPE_CHECKING:  # pragma: no cover
    from cashews._typing import TTL, DecoratedFunc, Exceptions, Key, KeyOrTemplate


def circuit_breaker(
    backend: _BackendInterface,
    errors_rate: int,
    period: TTL,
    ttl: TTL,
    half_open_ttl: TTL | None = None,
    min_calls: int = 1,
    exceptions: Exceptions = Exception,
    key: KeyOrTemplate | None = None,
    prefix: str = "circuit_breaker",
) -> Callable[[DecoratedFunc], DecoratedFunc]:
    """
    Circuit breaker
    :param backend: cache backend
    :param errors_rate: Errors rate in percents
    :param period: Period
    :param ttl: seconds in int or as timedelta to keep circuit breaker switched
    :param min_calls: numbers of call before circuit breaker can switch
    :param exceptions: exceptions at which returned cache result
    :param key: custom cache key, may contain alias to args or kwargs passed to a call
    :param prefix: custom prefix for key, default "circuit_breaker"
    """
    ttl = ttl_to_seconds(ttl)
    period = ttl_to_seconds(period)
    half_open_ttl = ttl_to_seconds(half_open_ttl)
    assert 0 < errors_rate < 100

    def _decor(func: DecoratedFunc) -> DecoratedFunc:
        _key = ":".join([func.__module__, func.__name__])
        _key_template = get_cache_key_template(func, key=key or _key, prefix=prefix)

        @wraps(func)
        async def _wrap(*args, **kwargs):
            _cache_key = get_cache_key(func, _key_template, args, kwargs)
            if await backend.is_locked(_cache_key + ":open"):
                if half_open_ttl:
                    await backend.set(
                        _cache_key + ":halfopen",
                        value=1,
                        expire=half_open_ttl,
                        exist=False,
                    )
                raise CircuitBreakerOpen()
            if await backend.exists(_cache_key + ":halfopen") and random.randint(0, 1):
                raise CircuitBreakerOpen()
            total = await _get_requests_count(backend, _cache_key + ":total", period)
            try:
                return await func(*args, **kwargs)
            except exceptions:
                fails = await _get_requests_count(backend, _cache_key + ":fails", period)
                if total and not total < min_calls and fails * 100 / total >= errors_rate:
                    await backend.set_lock(_cache_key + ":open", value=1, expire=ttl)
                raise

        return _wrap  # type: ignore[return-value]

    return _decor


async def _get_requests_count(backend: _BackendInterface, key: Key, period: int) -> int:
    timestamp = datetime.now(timezone.utc).timestamp()
    return await backend.slice_incr(key, timestamp - period, timestamp, 9999, expire=period)



import asyncio
from unittest.mock import ANY, Mock, PropertyMock

import pytest

from cashews.backends.memory import Memory
from cashews.cache_condition import NOT_NONE
from cashews.exceptions import NotConfiguredError
from cashews.wrapper import Cache
from cashews.wrapper.auto_init import create_auto_init


async def test_prefix_many(cache):
    await cache.init("mem://")
    await cache.init("mem://", prefix="-")

    await cache.set("key", "value")
    await cache.set("-:key", "-value")

    assert await cache.get("key") == "value"
    assert await cache.get("-:key") == "-value"

    assert await cache.get_many("key", "-:key") == ("value", "-value")


async def test_init():
    cache = Cache()
    assert cache.is_init
    cache.setup("mem://localhost", prefix="test")
    assert not cache.is_init

    await cache.init("mem://localhost", prefix="new")
    assert cache.is_init

    cache.setup("mem://localhost")
    assert not cache.is_init
    await cache.close()


async def test_auto_init(cache):
    target = Mock(wraps=Memory())
    init = False

    def set_init():
        async def _set():
            nonlocal init
            await asyncio.sleep(0.01)
            init = True

        return _set()

    type(target).is_init = PropertyMock(side_effect=lambda: init)
    target.init.side_effect = set_init
    cache._backends[""] = target
    cache._middlewares[target._id] = [create_auto_init()]

    await asyncio.gather(cache.ping(), cache.ping(), cache.get("test"))
    target.init.assert_called_once()


async def test_smoke_cmds(cache: Cache, target: Mock):
    await cache.set(key="key", value={"any": True}, expire=60, exist=None)
    target.set.assert_called_once_with(
        key="key",
        value=ANY,
        expire=60,
        exist=None,
    )

    await cache.set_raw(key="key2", value="value")
    target.set_raw.assert_called_once_with(key="key2", value="value")

    await cache.get("key")  # -> Any
    target.get.assert_called_once_with(key="key", default=None)

    await cache.get_raw("key")  # -> Any
    target.get_raw.assert_called_once_with(key="key")

    await cache.set_many({"key1": "value1", "key2": "value2"}, expire=60)
    target.set_many.assert_called_once_with(pairs={"key1": ANY, "key2": ANY}, expire=60)

    await cache.get_many("key1", "key2")
    target.get_many.assert_called_once_with("key1", "key2", default=None)

    await cache.incr("key_incr")  # -> int
    target.incr.assert_called_once_with(key="key_incr", value=1, expire=None)

    await cache.delete("key")
    target.delete.assert_called_once_with(key="key")

    await cache.delete_match("key*")
    target.delete_match.assert_called_once_with(pattern="key*")

    await cache.delete_many("key", "key2")
    target.delete_many.assert_called_once_with("key", "key2")

    await cache.expire(key="key", timeout=10)
    target.expire.assert_called_once_with(key="key", timeout=10)

    await cache.get_expire(key="key")  # -> int seconds to expire
    target.get_expire.assert_called_once()

    await cache.ping(message=b"test")  # -> bytes
    target.ping.assert_called_once_with(message=b"test")

    await cache.clear()
    target.clear.assert_called_once_with()

    await cache.is_locked("key", wait=60)  # -> bool
    target.is_locked.assert_called_once_with(key="key", wait=60, step=0.1)

    await cache.set_lock("key", "value", expire=60)  # -> bool
    target.set_lock.assert_called_once_with(key="key", value="value", expire=60)

    await cache.unlock("key", "value")  # -> bool
    target.unlock.assert_called_once_with(key="key", value="value")

    await cache.exists("key")
    target.exists.assert_called_once_with(key="key")

    await cache.get_bits("bits_key", 1, 2, 3, size=2)
    target.get_bits.assert_called_once_with("bits_key", 1, 2, 3, size=2)

    await cache.incr_bits("bits_key", 1, 2, 3, size=2)
    target.incr_bits.assert_called_once_with("bits_key", 1, 2, 3, size=2, by=1)

    await cache.set("key", "value")
    assert [key async for key in cache.scan("key*")] == ["key"]
    target.scan.assert_called_once_with(pattern="key*", batch_size=100)

    assert [key_value async for key_value in cache.get_match("key*")] == [("key", "value")]
    target.get_match.assert_called_once_with(pattern="key*", batch_size=100)

    await cache.get_size("key")
    target.get_size.assert_called_once_with(key="key")

    await cache.slice_incr("key_slice", 0, 10, maxvalue=10)
    target.slice_incr.assert_called_once_with(key="key_slice", start=0, end=10, maxvalue=10, expire=None)

    await cache.set_pop("key_set", count=10)
    target.set_pop.assert_called_once_with(key="key_set", count=10)

    await cache.set_add("key_set", "val1", "val2")
    target.set_add.assert_called_once_with("key_set", "val1", "val2", expire=None)

    await cache.close()


async def test_disable_cache_on_fail_return(cache: Cache):
    assert await cache.get("key") is None

    @cache(ttl=1, key="key", upper=True)
    @cache.failover(ttl=1, key="fail", prefix="")
    async def func(fail=False):
        if fail:
            raise Exception()
        return "1"

    await cache.set("fail", "test_fail")

    assert await func(fail=True) == "test_fail"  # return from fail cache but simple cache should be skipped
    assert await cache.get("key") is None

    await cache.delete("fail")
    await cache.set("key", "val")
    assert await func() == "val"


async def test_disable_cache_on_fail_return_2(cache: Cache):
    assert await cache.get("key") is None

    @cache.failover(ttl=1, key="fail", prefix="")
    @cache(ttl=1, key="key")
    async def func(fail=False):
        if fail:
            raise Exception()
        return "1"

    await cache.set("fail", "test_fail")

    assert await func(fail=True) == "test_fail"  # return from fail cache but simple cache should be skipped
    assert await cache.get("key") is None

    await cache.delete("fail")
    assert await cache.get("fail") is None
    await cache.set("key", "val")
    assert await func() == "val"
    assert await cache.get("fail") is None


async def test_multilayer_cache(cache: Cache):
    # If results from key2, key1 must not be set

    @cache(ttl=1, key="key1", upper=True, lock=True)
    @cache(ttl=1, key="key2")
    async def func():
        return 1

    await cache.set("key2", "test2")

    assert await func() == "test2"
    assert await cache.get("key1") is None

    await cache.set("key1", "test1")
    assert await func() == "test1"
    assert await cache.get("key2") == "test2"


async def test_cache_lock(cache: Cache):
    m = Mock()

    @cache(ttl=3, lock=True, protected=False)
    async def my_func(val=1):
        await asyncio.sleep(0)  # for task switching
        m(val)
        return val

    await asyncio.gather(my_func(), my_func(), my_func())

    m.assert_called_once_with(1)


_cache = Cache()
_cache.setup("mem://")


@pytest.mark.parametrize("decorator", (_cache, _cache.early, _cache.soft))
async def test_time_condition(decorator):
    m = Mock()

    @decorator(key="test", ttl=10, time_condition=0.1, protected=False)
    async def my_func(sleep=0.01):
        await asyncio.sleep(sleep)
        m()
        return sleep

    await my_func()
    await asyncio.gather(my_func(), my_func(), my_func())
    assert m.call_count == 4
    m.reset_mock()

    await asyncio.gather(my_func(0.15), my_func(), my_func(0.1))
    assert m.call_count == 3
    await my_func(0.15)
    await my_func()
    assert m.call_count == 3

    m.reset_mock()
    await _cache.clear()
    await my_func()
    assert m.call_count == 1

    await _cache.close()


async def test_cache_simple_not_none(cache: Cache):
    mock = Mock()

    @cache(ttl=0.1, key="key", condition=NOT_NONE)
    async def func():
        mock()
        return None

    assert await func() is None
    assert mock.call_count == 1

    assert await func() is None
    assert mock.call_count == 2


async def test_no_setup():
    cache = Cache()

    with pytest.raises(NotConfiguredError):
        await cache.get("test")


async def test_no_setup_decor():
    cache = Cache()

    @cache(ttl=0.1, key="key")
    async def func():
        return None

    with pytest.raises(NotConfiguredError):
        await func()



from datetime import timedelta

import pytest

from cashews.exceptions import WrongKeyError
from cashews.formatter import default_formatter
from cashews.key import get_cache_key, get_cache_key_template
from cashews.key_context import context as key_context
from cashews.key_context import register as register_context
from cashews.ttl import ttl_to_seconds


async def func1(a): ...


async def func2(a, *, k=None, **kwargs): ...


async def func3(a, *, k="test"): ...


class Klass:
    data = {"test": 1}

    def method(self, a, k=None):
        return a


@default_formatter.type_format(Klass)
def _some(value: Klass):
    return "hoho"


@default_formatter.register("call_method", preformat=False)
def call_method(value: Klass, a):
    return value.method(a)


def test_cache_func_key_dict():
    async def func(user): ...

    obj = type("user", (), {"name": "test", "one": True})()

    assert get_cache_key(func, template="test_key:{user.name}", args=(obj,)) == "test_key:test"

    obj = type("user", (), {"name": "new", "one": True})()
    assert get_cache_key(func, template="test_key:{user.name}", args=(obj,)) == "test_key:new"


@pytest.mark.parametrize(
    ("args", "kwargs", "template", "key"),
    (
        (
            ("A1", "a2", "a3"),
            {"kwarg1": "k1", "kwarg3": True},
            "{arg1}:{kwarg1}-{kwarg3}",
            "A1:k1-true",
        ),
        (
            (),
            {"kwarg1": "!@#$%^&*()"},
            "!@#$%^&*():{kwarg1}",
            "!@#$%^&*():!@#$%^&*()",
        ),
        (
            ("A1", "a2", "a3"),
            {"kwarg1": "k1", "kwarg0": True},
            None,
            "tests.test_key:func:arg1:A1:arg2:a2:a3:kwarg1:k1:kwarg2:true:kwarg0:true",
        ),
        (
            ("A1", "a2", "b", True, 1, 1.2, ("a", 1, 1.2)),
            {"kwarg1": "k1", "e": True, "d": 1},
            "{__args__}:{__kwargs__}",
            "b:true:1:1.2:a:1:1.2:d:1:e:true",
        ),
        (
            (),
            {"kwarg1": "k1", "e": True, "d": 1},
            "{kwarg1}:{__kwargs__}",
            "k1:d:1:e:true",
        ),
        (
            (b"a1", "a2", "a3"),
            None,
            "{arg1}-{kwarg1}-{kwarg3}",
            "a1--",
        ),
        (
            (True, "a2", "a3"),
            None,
            "{arg1}-{kwarg1}-{kwarg3}",
            "true--",
        ),
        (
            (b"\x16F\xbd\xb0\xcf\xcdN\xd7Y)\xfa\x1d\x96\xb1u\x81", ""),
            None,
            "{arg1}",
            "1646bdb0cfcd4ed75929fa1d96b17581",
        ),
        (
            (Klass(), ""),
            None,
            "{arg1}",
            "hoho",
        ),
        (
            ("a1",),
            {"arg2": 2, "kwarg1": "k1"},
            "{arg2}-{kwarg1}-{kwarg3}",
            "2-k1-",
        ),
        (
            ("a1",),
            {"arg2": 2, "kwarg1": "k1", "kwarg3": "k3"},
            "{arg2}:{kwarg1}:{kwarg3}",
            "2:k1:k3",
        ),
        (
            ("a1",),
            {"kwarg1": "K1", "arg2": 2},
            "{arg2}:{kwarg1}:{kwarg3}",
            "2:K1:",
        ),
        (("a1", "a2"), {"kwarg1": 1234}, "{kwarg1:len}", "4"),
        (
            ("a1", "a2"),
            {"user": type("user", (), {"name": "test"})()},
            "{user.name:len}",
            "4",
        ),
        (
            ("a1", "a2"),
            {"kwarg1": "test"},
            "{kwarg1:hash}",
            "098f6bcd4621d373cade4e832627b4f6",
        ),
        (
            ("a1", "a2"),
            {"kwarg1": "test"},
            "{kwarg1:hash(md5)}",
            "098f6bcd4621d373cade4e832627b4f6",
        ),
        (
            ("a1", "a2"),
            {"kwarg1": "test", "kwarg2": "md5"},
            "{kwarg1:hash({kwarg2})}",
            "098f6bcd4621d373cade4e832627b4f6",
        ),
        (
            ("a1", "a2"),
            {"kwarg1": Klass(), "kwarg2": "expect"},
            "{kwarg1:call_method({kwarg2})}",
            "expect",
        ),
        (
            ("a1", "a2"),
            {"kwarg1": "eyJhbGciOiJIUzI1NiJ9.eyJ1c2VyIjoidGVzdCJ9.n75bSCIEuSemX5iop0sCF3HmXwMQWF1zI6TzckzHUKY"},
            "{kwarg1:jwt(user)}",
            "test",
        ),
        (
            (),
            {"kwarg": "1"},
            "{@:get(context_value)}:{kwarg}",
            "context:1",
        ),
        (
            (),
            {"kwarg": 1},
            "{@:get(context_value)}:{kwarg}",
            "context:1",
        ),
        (
            (),
            {"kwarg": {1: 1}},
            "{@:get(context_value)}:{kwarg}",
            "context:1:1",
        ),
    ),
)
def test_cache_key_args_kwargs(args, kwargs, template, key):
    async def func(arg1, arg2, *args, kwarg1=None, kwarg2=b"true", **kwargs): ...

    with key_context(context_value="context", kwarg1="test"):
        assert get_cache_key(func, template, args=args, kwargs=kwargs) == key


@pytest.mark.parametrize(
    ("func", "key", "template"),
    (
        (func1, None, "tests.test_key:func1:a:{a}"),
        (func2, None, "tests.test_key:func2:a:{a}:k:{k}:{__kwargs__}"),
        (func3, None, "tests.test_key:func3:a:{a}:k:{k}"),
        (Klass.method, None, "tests.test_key:Klass.method:self:{self}:a:{a}:k:{k}"),
        (Klass.method, "key:{k}:{self.data.test}", "key:{k}:{self.data.test}"),
        (func2, "key:{k}", "key:{k}"),
        (func3, "key:{k:len}:{k:hash(md5)}:{val}", "key:{k:len}:{k:hash(md5)}:{val}"),
        (func3, "key:{k:len}:{k:hash(md5)}:{@:get(val)}", "key:{k:len}:{k:hash(md5)}:{@:get(val)}"),
    ),
)
def test_get_key_template(func, key, template):
    register_context("val", "k")
    assert get_cache_key_template(func, key) == template


def test_get_key_template_error():
    with pytest.raises(WrongKeyError) as exc:
        get_cache_key_template(func1, "key:{wrong_key}:{a}")
    exc.match("wrong_key")


@pytest.mark.parametrize(
    ("ttl", "expect"),
    (
        (timedelta(seconds=10), 10),
        (10, 10),
        (100.1, 100.1),
        ("10s", 10),
        ("1m10s", 60 + 10),
        ("10m1s", 60 * 10 + 1),
        ("1", 1),
        ("80", 80),
    ),
)
def test_ttl_to_seconds(ttl, expect):
    assert ttl_to_seconds(ttl) == expect



import asyncio
from decimal import Decimal

import pytest

from cashews import Cache
from cashews.backends.interface import NOT_EXIST, UNLIMITED
from cashews.backends.memory import Memory

VALUE = Decimal("100.2")


async def test_set_get(cache: Cache):
    await cache.set("key", VALUE)
    assert await cache.get("key") == VALUE


async def test_get_or_set(cache: Cache):
    await cache.get_or_set("key", default=VALUE)
    assert await cache.get("key") == VALUE


async def test_get_or_set_callable(cache: Cache):
    await cache.get_or_set("key", default=lambda: VALUE)
    assert await cache.get("key") == VALUE


async def test_get_or_set_awaitable(cache: Cache):
    async def _default():
        return VALUE

    await cache.get_or_set("key", default=_default)
    assert await cache.get("key") == VALUE


async def test_get_or_set_coroutine(cache: Cache):
    async def _default():
        return VALUE

    await cache.get_or_set("key", default=_default())
    assert await cache.get("key") == VALUE


async def test_get_or_set_no_set(cache: Cache):
    await cache.set("key", VALUE)
    await cache.get_or_set("key", None)
    assert await cache.get("key") == VALUE


async def test_set_get_bytes(cache: Cache):
    await cache.set("key", b"10")
    assert await cache.get("key") == b"10"


async def test_mutation(cache):
    obj = [1, 2, 3]
    await cache.set("key", obj)
    obj.append(4)
    assert list(await cache.get("key")) != list(obj)
    assert await cache.get("key") == [1, 2, 3]


async def test_incr_get(cache: Cache):
    await cache.incr("key")
    assert await cache.get("key") == 1


async def test_incr_value_get(cache: Cache):
    await cache.incr("key")
    await cache.incr("key", 2)
    assert await cache.get("key") == 3


async def test_incr_expire(cache: Cache):
    await cache.incr("key", expire=10)
    assert await cache.get_expire("key") == 10


async def test_set_get_many(cache: Cache):
    await cache.set("key", VALUE)
    assert await cache.get_many("key", "no_exists") == (VALUE, None)


async def test_diff_types_get_many(cache: Cache):
    await cache.incr("key")
    await cache.incr_bits("key2", 0)
    assert await cache.get_many("key", "key2") == (1, None)


async def test_set_exist(cache: Cache):
    assert await cache.set("key", "value")
    assert await cache.set("key", VALUE, exist=True)
    assert await cache.get("key") == VALUE

    assert not await cache.set("key2", "value", exist=True)
    assert await cache.get("key2") is None

    assert await cache.set("key2", "value", exist=False)
    assert await cache.get("key2") == "value"

    assert not await cache.set("key2", "value2", exist=False)
    assert await cache.get("key2") == "value"


async def test_set_many(cache: Cache):
    await cache.set_many({"key1": VALUE, "key2": "value2"}, expire=1)
    assert await cache.get("key1", VALUE)
    assert await cache.get("key2", "value2")


async def test_get_no_value(cache: Cache):
    assert await cache.get("key2") is None
    assert await cache.get("key2", default=VALUE) is VALUE


async def test_incr(cache: Cache):
    assert await cache.incr("incr") == 1
    assert await cache.incr("incr") == 2
    assert await cache.incr("incr", 5) == 7
    assert await cache.get("incr") == 7


async def test_incr_set(cache: Cache):
    await cache.set("incr", "test")
    with pytest.raises(Exception):
        assert await cache.incr("incr") == 1
    assert await cache.get("incr") == "test"


async def test_ping(cache: Cache):
    assert await cache.ping() == b"PONG"


async def test_exists(cache: Cache):
    assert not await cache.exists("not")
    await cache.set("yes", VALUE)
    assert await cache.exists("yes")


async def test_expire(cache: Cache):
    await cache.set("key", VALUE, expire=0.01)
    assert await cache.get("key") == VALUE
    await asyncio.sleep(0.01)
    assert await cache.get("key") is None


async def test_get_set_expire(cache: Cache):
    assert await cache.get_expire("key") == NOT_EXIST
    await cache.set("key", VALUE)
    assert await cache.get("key") == VALUE
    assert await cache.get_expire("key") == UNLIMITED
    await cache.expire("key", 1)
    assert await cache.get_expire("key") == 1


async def test_delete_many(cache: Cache):
    await cache.set("key1", VALUE)
    await cache.set("key2", VALUE)
    await cache.set("key3", VALUE)

    await cache.delete_many("key1", "key2", "key4")

    assert await cache.get("key1") is None
    assert await cache.get("key2") is None
    assert await cache.get("key3") == VALUE


async def test_delete_match(cache: Cache):
    await cache.set("pref:test:test", VALUE)
    await cache.set("pref:value:test", b"value2")
    await cache.set("pref:-:test", b"-")
    await cache.set("pref:*:test", b"*")

    await cache.set("ppref:test:test", b"value3")
    await cache.set("pref:test:tests", b"value3")

    await cache.delete_match("pref:*:test")

    assert await cache.get("pref:test:test") is None
    assert await cache.get("pref:value:test") is None
    assert await cache.get("pref:-:test") is None
    assert await cache.get("pref:*:test") is None

    assert await cache.get("ppref:test:test") is not None
    assert await cache.get("pref:test:tests") is not None


async def test_scan(cache: Cache):
    await cache.set("pref:test:test", VALUE)
    await cache.set("pref:value:test", b"value2")
    await cache.set("pref:-:test", b"-")
    await cache.set("pref:*:test", b"*")

    await cache.set("ppref:test:test", b"value3")
    await cache.set("pref:test:tests", b"value3")

    keys = [key async for key in cache.scan("pref:*:test")]

    assert len(keys) == 4
    assert set(keys) == {"pref:test:test", "pref:value:test", "pref:-:test", "pref:*:test"}


async def test_get_match(cache: Cache):
    await cache.set("pref:test:test", VALUE)
    await cache.set("pref:value:test", b"value2")
    await cache.set("pref:-:test", b"-")
    await cache.set("pref:*:test", b"*")

    await cache.set("ppref:test:test", b"value3")
    await cache.set("pref:test:tests", b"value3")

    match = [(key, value) async for key, value in cache.get_match("pref:*:test")]

    assert len(match) == 4
    assert dict(match) == {
        "pref:test:test": VALUE,
        "pref:value:test": b"value2",
        "pref:-:test": b"-",
        "pref:*:test": b"*",
    }
    match = [(key, value) async for key, value in cache.get_match("not_exists:*")]
    assert len(match) == 0


async def test_set_lock_unlock(cache: Cache):
    await cache.set_lock("lock", "lock", 10)
    assert await cache.is_locked("lock")
    await cache.unlock("lock", "lock")
    assert not await cache.is_locked("lock")


async def test_lock(cache: Cache):
    async with cache.lock("lock", 10):
        assert await cache.is_locked("lock")

    assert not await cache.is_locked("lock")


async def test_diff_types_get_match(cache: Cache):
    await cache.incr("key")
    await cache.incr_bits("key2", 0)
    match = [(key, value) async for key, value in cache.get_match("*")]
    assert len(match) == 1, match
    assert dict(match) == {"key": 1}


async def test_get_size(cache: Cache):
    await cache.set("test", b"1")
    assert isinstance(await cache.get_size("test"), int)


async def test_get_keys_count(cache: Cache):
    await cache.set("test", b"1")
    assert await cache.get_keys_count() == 1


async def test_get_bits(cache: Cache):
    assert await cache.get_bits("test", 0, 2, 10, 50000, size=1) == (0, 0, 0, 0)
    assert await cache.get_bits("test", 0, 1, 3, size=15) == (
        0,
        0,
        0,
    )


async def test_incr_bits(cache: Cache):
    await cache.incr_bits("test", 0, 1, 4)
    assert await cache.get_bits("test", 0, 1, 2, 3, 4) == (1, 1, 0, 0, 1)


async def test_bits_size(cache: Cache):
    await cache.incr_bits("test", 0, 1, 4, size=5, by=3)
    assert await cache.get_bits("test", 0, 1, 2, 3, 4, size=5) == (3, 3, 0, 0, 3)


async def test_slice_incr(cache: Cache):
    assert await cache.slice_incr("test", 0, 5, maxvalue=10) == 1
    assert await cache.slice_incr("test", 1, 6, maxvalue=10) == 2
    assert await cache.slice_incr("test", 2, 7, maxvalue=10) == 3
    assert await cache.slice_incr("test", 3, 8, maxvalue=10) == 4
    assert await cache.slice_incr("test", 5, 10, maxvalue=4) == 4
    assert await cache.slice_incr("test", 9, 11, maxvalue=10) == 1
    assert await cache.slice_incr("test", 15, 20, maxvalue=10) == 1
    assert await cache.slice_incr("test", 9, 11, 10) == 1


async def test_lru(backend_factory):
    cache = backend_factory(Memory, size=10)
    # fill cache
    for i in range(10):
        await cache.set(f"key:{i}", i)

    # use only 5 first
    for i in range(5):
        await cache.get(f"key:{i}")

    # add 5 more keys
    for i in range(5):
        await cache.set(f"key:{i}:new", i)

    assert len(cache.store) == 10

    for i in range(5):
        assert await cache.get(f"key:{i}") == i

    for i in range(6, 10):
        assert await cache.get(f"key:{i}") is None

    await cache.close()


async def test_lru2(backend_factory):
    cache = backend_factory(Memory, size=10)
    # fill cache
    for i in range(10):
        await cache.set(f"key:{i}", i)

    # use only 5 last
    for i in range(6, 10):
        await cache.get(f"key:{i}")

    # add 5 more keys
    for i in range(5):
        await cache.set(f"key:{i}:new", i)

    assert len(cache.store) == 10

    for i in range(5):
        assert await cache.get(f"key:{i}") is None

    for i in range(6, 10):
        assert await cache.get(f"key:{i}") == i

    await cache.close()



from __future__ import annotations

import asyncio
import re
import time
from collections import OrderedDict
from collections.abc import AsyncIterator, Iterable, Mapping
from contextlib import suppress
from copy import copy
from typing import TYPE_CHECKING, Any, overload

from cashews.utils import Bitarray, get_obj_size

from .interface import NOT_EXIST, UNLIMITED, Backend

if TYPE_CHECKING:  # pragma: no cover
    from cashews._typing import Default, Key, Value


__all__ = ["Memory"]

_missed = object()


class Memory(Backend):
    """
    Inmemory backend lru with ttl
    """

    __slots__ = [
        "store",
        "_check_interval",
        "size",
        "__is_init",
        "__remove_expired_stop",
        "__remove_expired_task",
    ]

    def __init__(self, size: int = 1000, check_interval: float = 1, **kwargs: Any) -> None:
        self.store: OrderedDict[Key, tuple[float | None, Value]] = OrderedDict()
        self._check_interval = check_interval
        self.size = size
        self.__is_init = False
        self.__remove_expired_stop: asyncio.Event = asyncio.Event()
        self.__remove_expired_task: asyncio.Task[None] | None = None
        super().__init__(**kwargs)

    async def init(self) -> None:
        self.__is_init = True
        if self._check_interval:
            self.__remove_expired_stop = asyncio.Event()
            self.__remove_expired_task = asyncio.create_task(self._remove_expired())

    @property
    def is_init(self) -> bool:
        return self.__is_init

    async def _remove_expired(self) -> None:
        while not self.__remove_expired_stop.is_set():
            for key in dict(self.store):
                await self.get(key)
            with suppress(asyncio.TimeoutError, TimeoutError):
                await asyncio.wait_for(self.__remove_expired_stop.wait(), self._check_interval)

    async def clear(self) -> None:
        self.store = OrderedDict()

    async def set(
        self,
        key: Key,
        value: Value,
        expire: float | None = None,
        exist: bool | None = None,
    ) -> bool:
        if exist is not None and (key in self.store) is not exist:
            return False
        if self._serializer:
            value = await self._serializer.encode(self, key=key, value=value, expire=expire)
        self._set(key, value, expire)
        return True

    async def set_raw(self, key: Key, value: Value, **kwargs: Any) -> None:
        self.store[key] = (None, value)

    async def get(self, key: Key, default: Value | None = None) -> Value:
        return await self._get(key, default=default)

    async def get_raw(self, key: Key) -> Value | None:
        val = self.store.get(key)
        if val:
            return val[1]
        return None

    async def get_many(self, *keys: Key, default: Value | None = None) -> tuple[Value | None, ...]:
        values = []
        for key in keys:
            val = await self._get(key, default=default)
            if isinstance(val, Bitarray):
                continue
            values.append(val)
        return tuple(values)

    async def set_many(self, pairs: Mapping[Key, Value], expire: float | None = None) -> None:
        for key, value in pairs.items():
            if self._serializer:
                value = await self._serializer.encode(self, key=key, value=value, expire=expire)
            self._set(key, value, expire)

    async def scan(self, pattern: str, batch_size: int = 100) -> AsyncIterator[Key]:
        pattern = pattern.replace("*", ".*")
        regexp = re.compile(pattern)
        for key in dict(self.store):
            if regexp.fullmatch(key):
                yield key

    async def incr(self, key: Key, value: int = 1, expire: float | None = None) -> int:
        value += int(await self._get(key, 0))
        _expire = None if value != 1 else expire
        self._set(key=key, value=value, expire=_expire)
        return value

    async def exists(self, key: Key) -> bool:
        return await self._key_exist(key)

    async def delete(self, key: Key) -> bool:
        return await self._delete(key)

    async def _delete(self, key: Key) -> bool:
        if key in self.store:
            del self.store[key]
            await self._call_on_remove_callbacks(key)
            return True
        return False

    async def delete_many(self, *keys: Key) -> None:
        for key in keys:
            await self._delete(key)

    async def delete_match(self, pattern: Key) -> None:
        async for key in self.scan(pattern):
            await self._delete(key)

    async def get_match(
        self,
        pattern: str,
        batch_size: int = 100,
    ) -> AsyncIterator[tuple[Key, Value]]:
        async for key in self.scan(pattern):
            value = await self.get(key)
            if not isinstance(value, Bitarray):
                yield key, value

    async def expire(self, key: Key, timeout: float) -> None:
        if not await self._key_exist(key):
            return
        value = await self._get(key, default=_missed)
        if value is _missed:
            return
        self._set(key, value, timeout)

    async def get_expire(self, key: Key) -> int:
        if key not in self.store:
            return NOT_EXIST
        expire_at, _ = self.store[key]
        if expire_at is not None:
            return round(expire_at - time.time())
        return UNLIMITED

    async def ping(self, message: bytes | None = None) -> bytes:
        if message is None or message == b"PING":
            return b"PONG"
        return message

    async def get_bits(self, key: Key, *indexes: int, size: int = 1) -> tuple[int, ...]:
        array: Bitarray = await self._get(key, default=Bitarray("0"))
        return tuple(array.get(index, size) for index in indexes)

    async def incr_bits(self, key: Key, *indexes: int, size: int = 1, by: int = 1) -> tuple[int, ...]:
        array: Bitarray = await self._get(key, default=Bitarray("0"))
        result = []
        for index in indexes:
            array.incr(index, size, by)
            result.append(array.get(index, size))
        self._set(key, array)
        return tuple(result)

    def _set(self, key: Key, value: Value, expire: float | None = None) -> None:
        expire_time: float | None = time.time() + expire if expire else None
        if expire_time is None and key in self.store:
            expire_time, _ = self.store[key]
        self.store[key] = (expire_time, copy(value))
        self.store.move_to_end(key)
        if len(self.store) > self.size:
            self.store.popitem(last=False)

    @overload
    async def _get(self, key: Key, default: Default) -> Value | Default: ...

    @overload
    async def _get(self, key: Key, default: None = None) -> Value | None: ...

    async def _get(self, key: Key, default: Default | None = None) -> Value | None:
        if key not in self.store:
            return default
        self.store.move_to_end(key)
        expire_at, value = self.store[key]
        if expire_at and expire_at < time.time():
            await self._delete(key)
            return default
        if not self._serializer:
            return value
        return await self._serializer.decode(self, key=key, value=value, default=default)

    async def _key_exist(self, key: Key) -> bool:
        return (await self._get(key, default=_missed)) is not _missed

    async def is_locked(self, key: Key, wait: float | None = None, step: float = 0.1) -> bool:
        if wait is None:
            return await self._key_exist(key)
        while wait > 0:
            if not await self._key_exist(key):
                return False
            wait -= step
            await asyncio.sleep(step)
        return await self._key_exist(key)

    async def unlock(self, key: Key, value: Value) -> bool:
        return await self._delete(key)

    async def get_size(self, key: Key) -> int:
        if key in self.store:
            return get_obj_size(self.store[key])
        return 0

    async def slice_incr(
        self,
        key: Key,
        start: int | float,
        end: int | float,
        maxvalue: int,
        expire: float | None = None,
    ) -> int:
        val_list = await self._get(key)
        count = 0
        new_val = []
        if val_list:
            for val in val_list:
                if start <= val < end:
                    count += 1
                    new_val.append(val)
        if count < maxvalue:
            count += 1
            new_val.append(end)
        self._set(key, new_val, expire=expire)
        return count

    async def set_add(self, key: Key, *values: str, expire: float | None = None) -> None:
        val: set[str] = await self._get(key, default=set())
        val.update(values)
        self._set(key, val, expire=expire)

    async def set_remove(self, key: Key, *values: str) -> None:
        val: set[str] = await self._get(key, default=set())
        val.difference_update(values)
        self._set(key, val)

    async def set_pop(self, key: Key, count: int = 100) -> Iterable[str]:
        values: set[str] = await self._get(key, default=set())
        _values: list[str] = []
        for _ in range(count):
            if not values:
                break
            _values.append(values.pop())

        self._set(key, values)
        return _values

    async def get_keys_count(self) -> int:
        return len(self.store)

    async def close(self) -> None:
        if self.__remove_expired_task:
            self.__remove_expired_stop.set()
            await self.__remove_expired_task
            del self.__remove_expired_task
            self.__remove_expired_task = None
        if self.__remove_expired_stop:
            del self.__remove_expired_stop
            self.__remove_expired_stop = None  # type: ignore[assignment]
        self.__is_init = False



from unittest.mock import Mock

import pytest

from cashews import CacheBackendInteractionError
from cashews.wrapper import Cache

pytestmark = [pytest.mark.asyncio, pytest.mark.redis]


@pytest.fixture
def redis_backend():
    from cashews.backends.redis import Redis

    return Redis


async def test_safe_redis(redis_backend):
    redis = redis_backend(suppress=True, address="redis://localhost:9223")
    await redis.init()

    assert await redis.set("test", "test") is False
    assert await redis.set_raw("test", "test") is None

    assert await redis.set_lock("test", "test", 1) is False
    assert await redis.unlock("test", "test") is None
    assert await redis.is_locked("test") is False

    assert await redis.get("test", default="no") == "no"
    assert await redis.get("test") is None
    assert await redis.get_raw("test") is None
    assert await redis.get_many("test", "test2") == (None, None)

    await redis.set_many({"test": "test", "test2": "test2"}, expire=1)  # failed - need to fix
    await redis.set_many({"test": "test", "test2": "test2"})

    assert await redis.get_expire("test") == 0
    assert await redis.incr("test") is None
    assert await redis.slice_incr("test", 1, 2, 10) is None

    async for _ in redis.scan("*"):
        raise AssertionError()

    async for _ in redis.get_match("*"):
        raise AssertionError()

    assert await redis.set_add("testset", "test") is None
    assert await redis.set_pop("testset") == []

    assert await redis.delete("test") == 0
    await redis.delete_match("*")

    with pytest.raises(CacheBackendInteractionError):
        await redis.ping()


async def test_unsafe_redis_down(redis_backend):
    redis = redis_backend(suppress=False, address="redis://localhost:9223")
    await redis.init()
    with pytest.raises(CacheBackendInteractionError):
        await redis.ping()
    with pytest.raises(CacheBackendInteractionError):
        await redis.set("key", "value")


async def test_cache_decorators_on_redis_down(redis_backend):
    mock = Mock(return_value="val")
    cache = Cache()
    cache._add_backend(redis_backend(suppress=True, address="redis://localhost:9223"))

    @cache(ttl=1)
    @cache.failover(1)
    @cache.hit(ttl=1, cache_hits=1)
    @cache.circuit_breaker(ttl=1, errors_rate=1, period=1)
    @cache.rate_limit(ttl=1, limit=1, period=1)
    @cache.early(ttl=1)
    @cache.dynamic()
    @cache.locked(ttl=1)
    async def func():
        return mock()

    assert await func() == "val"
    assert mock.call_count == 1

    assert await func() == "val"
    assert mock.call_count == 2



import asyncio
from random import random
from unittest.mock import Mock

from cashews.commands import Command
from cashews.wrapper import Cache


async def test_disable_cmd(cache):
    cache.disable(Command.INCR)
    await cache.set("test", 10)
    await cache.incr("test")
    assert await cache.get("test") == 10

    cache.enable(Command.INCR)
    await cache.incr("test")
    assert await cache.get("test") == 11


async def test_is_disable():
    cache = Cache()
    await cache.init("mem://localhost")

    assert not cache.is_disable()
    assert not cache.is_disable(Command.GET)
    assert not cache.is_full_disable

    assert cache.is_enable()
    assert cache.is_enable(Command.GET)

    cache.disable(Command.GET)

    assert cache.is_disable()
    assert cache.is_disable(Command.GET)
    assert not cache.is_disable(Command.SET)
    assert not cache.is_full_disable

    assert not cache.is_enable()
    assert not cache.is_enable(Command.GET)
    assert cache.is_enable(Command.SET)

    cache.disable()
    assert cache.is_disable()
    assert cache.is_disable(Command.GET)
    assert cache.is_disable(Command.SET)
    assert cache.is_full_disable

    assert not cache.is_enable()
    assert not cache.is_enable(Command.GET)
    assert not cache.is_enable(Command.SET)

    await cache.close()


async def test_disable_context_manager(cache):
    with cache.disabling(Command.INCR):
        await cache.set("test", 10)
        await cache.incr("test")
    assert await cache.get("test") == 10


async def test_disable_context_manage_get(cache):
    await cache.set("test", 10)
    with cache.disabling():
        assert await cache.get("test") is None


def test_disable_context_manage_no_init():
    cache = Cache()
    cache.disable(Command.GET)


async def test_disable_context_manage_decor(cache):
    @cache(ttl="1m")
    async def func():
        return random()

    was = await func()
    with cache.disabling():
        assert await func() != was


async def test_init_disable(cache):
    cache.setup("mem://localhost?disable=1")
    assert cache.is_disable()


async def test_init_enable(cache):
    cache.setup("mem://localhost?enable=1")
    assert cache.is_enable()
    assert not cache.is_disable()


async def test_prefix_cache(cache):
    await cache.set("-:key", "value")

    cache.setup("://", prefix="-", disable=True)
    assert not cache.is_disable()
    assert cache.is_disable(prefix="-")

    await cache.set("key", "value")
    await cache.set("-:key", "new")

    assert await cache.get("key") == "value"
    assert await cache.get("-:key") is None

    cache.enable(prefix="-")
    assert cache.is_enable(prefix="-")

    assert await cache.get("-:key") is None  # new backend haven't this key


async def test_disable_ctz(cache):
    cache.enable()

    async def test():
        await cache.set("test", "1")
        assert await cache.get("test") == "1"
        cache.disable(Command.SET)
        await cache.set("test", "2")

    await asyncio.create_task(test())
    assert await cache.get("test") == "1"
    await cache.set("test", "3")
    assert await cache.get("test") == "3"


async def test_disable_decorators(cache: Cache, target):
    cache.disable()
    data = (i for i in range(10))
    target.is_full_disable = False

    @cache(ttl=1)
    @cache.soft(ttl=1)
    @cache.failover(ttl=1)
    @cache.hit(ttl=1, cache_hits=1)
    @cache.circuit_breaker(ttl=1, errors_rate=1, period=1)
    @cache.rate_limit(ttl=1, limit=1, period=1)
    @cache.early(ttl=1)
    @cache.dynamic()
    @cache.locked(ttl=1)
    async def func():
        return next(data)

    assert await func() == 0
    assert await func() == 1
    target.get.assert_not_called()
    target.set.assert_not_called()

    cache.enable()
    assert await func() == 2
    assert await func() == 2


async def test_disable_bloom(cache: Cache, target: Mock):
    cache.disable()

    @cache.bloom(capacity=10)
    async def func():
        return True

    await func.set()
    assert await func()
    target.incr_bits.assert_not_called()
    target.get_bits.assert_not_called()

    cache.enable()
    await func.set()
    assert await func()
    target.incr_bits.assert_called()
    target.get_bits.assert_called()


async def test_disable_decorators_get(cache: Cache):
    data = (i for i in range(10))

    @cache(ttl=1)
    async def func():
        return next(data)

    assert cache.is_enable()
    assert cache.is_enable(Command.SET, prefix="cache")
    assert cache.is_enable(Command.GET, prefix="cache")
    assert cache.is_enable(Command.SET, prefix="")
    assert await func() == 0
    assert await func() == 0

    cache.disable(Command.GET)

    assert not cache.is_enable(Command.GET)
    assert cache.is_enable(Command.SET)

    assert await func() == 1
    assert await func() == 2

    cache.enable(Command.GET)
    assert await func() == 2


async def test_disable_decorator_set(cache):
    data = (i for i in range(10))
    cache.disable(Command.SET)

    @cache(ttl=1)
    async def func():
        return next(data)

    assert await func() == 0
    assert await func() == 1

    cache.enable(Command.SET)
    assert await func() == 2
    assert await func() == 2


async def test_disable_and_get_enable(cache):
    data = (i for i in range(10))
    cache.enable()
    assert cache.is_enable()
    assert not cache.is_full_disable

    @cache(ttl=1)
    async def func():
        return next(data)

    assert await func() == 0
    assert await func() == 0
    cache.disable()
    assert await func() == 1
    assert await func() == 2
    cache.enable(Command.GET)
    assert await func() == 0



from __future__ import annotations

from functools import wraps
from typing import TYPE_CHECKING, Callable

from cashews.key import get_cache_key, get_cache_key_template
from cashews.ttl import ttl_to_seconds

from ._exception import RaiseException, return_or_raise
from .defaults import _empty, context_cache_detect

if TYPE_CHECKING:  # pragma: no cover
    from cashews import Cache
    from cashews._typing import TTL, CallableCacheCondition, DecoratedFunc, KeyOrTemplate, Tags

__all__ = ("cache",)


def cache(
    backend: Cache,
    ttl: TTL,
    key: KeyOrTemplate | None = None,
    condition: CallableCacheCondition = lambda *args, **kwargs: True,
    prefix: str = "",
    tags: Tags = (),
) -> Callable[[DecoratedFunc], DecoratedFunc]:
    """
    Simple cache strategy - trying to return cached result,
    execute wrapped call and store a result with ttl if condition return true
    :param backend: cache backend
    :param ttl: duration in seconds to store a result or a callable
    :param key: custom cache key, may contain alias to args or kwargs passed to a call
    :param condition: callable object that determines whether the result will be saved or not
    :param prefix: custom prefix for key
    :param tags: aliases for keys that used for cache (used for invalidation)
    """

    ttl = ttl_to_seconds(ttl)

    def _decor(func: DecoratedFunc) -> DecoratedFunc:
        _key_template = get_cache_key_template(func, key=key, prefix=prefix)
        for tag in tags:
            backend.register_tag(tag, _key_template)

        @wraps(func)
        async def _wrap(*args, **kwargs):
            _tags = [get_cache_key(func, tag, args, kwargs) for tag in tags]
            _cache_key = get_cache_key(func, _key_template, args, kwargs)

            cached = await backend.get(_cache_key, default=_empty)
            if cached is not _empty:
                _ttl = ttl_to_seconds(ttl, *args, **kwargs, result=cached, with_callable=True)
                context_cache_detect._set(
                    _cache_key,
                    ttl=_ttl,
                    name="simple",
                    template=_key_template,
                    value=cached,
                )
                return return_or_raise(cached)
            _exc = None
            try:
                result = await func(*args, **kwargs)
            except Exception as exc:
                _exc = exc
                result = exc

            _ttl = ttl_to_seconds(ttl, *args, **kwargs, result=result, with_callable=True)
            cond_result = condition(result, args, kwargs, key=_cache_key)
            if isinstance(cond_result, bool) and cond_result and not isinstance(result, Exception):
                await backend.set(_cache_key, result, expire=_ttl, tags=_tags)
            elif isinstance(cond_result, Exception):
                await backend.set(_cache_key, RaiseException(result), expire=_ttl, tags=_tags)
            if _exc:
                raise _exc
            return result

        return _wrap  # type: ignore[return-value]

    return _decor



from __future__ import annotations

import asyncio
from collections.abc import AsyncIterator, Iterable, Mapping
from typing import Any

from redis.asyncio import BlockingConnectionPool
from redis.asyncio.client import Pipeline

from cashews._typing import Key, Value
from cashews.backends.interface import Backend
from cashews.serialize import DEFAULT_SERIALIZER, Serializer

from .client import Redis, SafePipeline, SafeRedis

_UNLOCK = """
if redis.call("GET", KEYS[1]) == ARGV[1] then
    return redis.call("DEL", KEYS[1])
else
    return 0
end
"""
_INCR_SLICE = """
redis.call("ZREMRANGEBYSCORE", KEYS[1], 0, "(" .. ARGV[1])
local current_count = redis.call("ZCOUNT", KEYS[1], ARGV[1], ARGV[2])
if current_count < tonumber(ARGV[3]) then
    current_count = current_count + 1
    redis.call("ZADD", KEYS[1], ARGV[2], ARGV[2])
    if tonumber(ARGV[4]) > 0 then
        redis.call("PEXPIRE", KEYS[1], ARGV[4])
    end
end
return current_count
"""
_INCR_EXPIRE = """
local current_count = redis.call("INCRBY", KEYS[1], ARGV[1])
if current_count == 1 then
    redis.call("PEXPIRE", KEYS[1], ARGV[2])
end
return current_count
"""
_empty = object()
# pylint: disable=arguments-differ
# pylint: disable=abstract-method


class _Redis(Backend):
    _client: Redis | SafeRedis
    _client_class: type[Redis] | type[SafeRedis]

    def __init__(
        self,
        address: str,
        suppress: bool = True,
        **kwargs: Any,
    ) -> None:
        kwargs.pop("local_cache", None)
        kwargs.pop("prefix", None)
        kwargs.setdefault("client_name", "cashews")
        kwargs.setdefault("health_check_interval", 10)
        kwargs.setdefault("max_connections", 10)
        kwargs.setdefault("retry_on_timeout", False)
        kwargs.setdefault("socket_timeout", 1)
        if not address.startswith("unix"):
            kwargs.setdefault("socket_keepalive", True)
        kwargs["decode_responses"] = False

        self._pool_class = kwargs.pop("connection_pool_class", BlockingConnectionPool)
        if self._pool_class == BlockingConnectionPool:
            kwargs["timeout"] = kwargs.pop("wait_for_connection_timeout", 10)
        self._sha: dict[str, Any] = {}
        if not suppress:
            self._client_class = Redis
            self._pipeline_class = Pipeline
        else:
            self._pipeline_class = SafePipeline
            self._client_class = SafeRedis
        self._kwargs = kwargs
        self._address = address
        self.__is_init = False
        super().__init__(serializer=kwargs.pop("serializer", None))
        self._serializer: Serializer = self._serializer or DEFAULT_SERIALIZER

    @property
    def is_init(self) -> bool:
        return self.__is_init

    async def init(self):
        pool = self._pool_class.from_url(self._address, **self._kwargs)
        if hasattr(self._client_class, "from_pool"):
            self._client = self._client_class.from_pool(pool)
        else:
            self._client = self._client_class(connection_pool=pool)
        await self._client.initialize()
        self.__is_init = True

    @property
    def _pipeline(self):
        return self._pipeline_class(self._client.connection_pool, self._client.response_callbacks, True, None)

    async def clear(self):
        return await self._client.flushdb()

    async def set(
        self,
        key: Key,
        value: Value,
        expire: float | None = None,
        exist=None,
    ) -> bool:
        value = await self._serializer.encode(self, key=key, value=value, expire=expire)
        nx = xx = False
        if exist is True:
            xx = True
        elif exist is False:
            nx = True
        px = int(expire * 1000) if expire else None
        _set = bool(await self._client.set(key, value, px=px, nx=nx, xx=xx))
        return _set

    async def set_many(self, pairs: Mapping[Key, Value], expire: float | None = None):
        px = int(expire * 1000) if expire else None
        async with self._pipeline as pipe:
            for key, value in pairs.items():
                value = await self._serializer.encode(self, key=key, value=value, expire=expire)
                await pipe.set(key, value, px=px)
            await pipe.execute()

    async def get_expire(self, key: Key) -> int:
        return await self._client.ttl(key)

    async def expire(self, key: Key, timeout: float):
        return await self._client.pexpire(key, int(timeout * 1000))

    async def set_lock(self, key: Key, value: Value, expire: float) -> bool:
        pexpire = int(expire * 1000)
        return bool(await self._client.set(key, value, px=pexpire, nx=True))

    async def is_locked(
        self,
        key: Key,
        wait: float | None = None,
        step: float = 0.1,
    ) -> bool:
        if wait is None:
            return await self.exists(key)
        while wait > 0.0:
            if not await self.exists(key):
                return False
            wait -= step
            await asyncio.sleep(step)
        return True

    async def unlock(self, key: Key, value: Value) -> bool:
        if self._sha.get("UNLOCK") is None:
            self._sha["UNLOCK"] = self._client.register_script(_UNLOCK.replace("\n", " "))
        return await self._sha["UNLOCK"](keys=(key,), args=(value,))

    async def delete(self, key: Key) -> bool:
        try:
            return bool(await self._client.unlink(key))
        finally:
            await self._call_on_remove_callbacks(key)

    async def exists(self, key: Key) -> bool:
        return bool(await self._client.exists(key))

    async def scan(self, pattern: str, batch_size: int = 100) -> AsyncIterator[Key]:
        cursor = 0
        while True:
            cursor, keys = await self._client.scan(cursor, match=pattern, count=batch_size)
            for key in keys:
                yield key.decode()
            if not cursor:
                return

    async def delete_many(self, *keys: Key):
        try:
            await self._client.unlink(*keys)
        finally:
            await self._call_on_remove_callbacks(*keys)

    async def delete_match(self, pattern: str):
        if "*" not in pattern:
            await self._client.unlink(pattern)
            return
        cursor = 0
        while True:
            cursor, keys = await self._client.scan(cursor, match=pattern, count=100)
            if not keys:
                if not cursor:
                    return
                continue
            await self._client.unlink(*keys)
            await self._call_on_remove_callbacks(*[key.decode() for key in keys])

    async def get_match(self, pattern: str, batch_size: int = 100) -> AsyncIterator[tuple[Key, Value]]:
        cursor = 0
        while True:
            cursor, keys = await self._client.scan(cursor, match=pattern, count=batch_size)
            if not keys:
                if not cursor:
                    return
                continue
            keys = [key.decode() for key in keys]
            values = await self.get_many(*keys, default=_empty)
            for key, value in zip(keys, values):
                if value is not _empty:  # key can be deleted after scan
                    yield key, value
            if not cursor:
                return

    async def get_size(self, key: Key) -> int:
        size = await self._client.memory_usage(key) or 0
        return int(size)

    async def get(self, key: Key, default: Value | None = None) -> Value:
        value = await self._client.get(key)
        return await self._transform_value(key, value, default)

    async def get_many(self, *keys: Key, default: Value | None = None) -> tuple[Value | None, ...]:
        if not keys:
            return ()
        values = await self._client.mget(*keys)
        if values is None:
            return tuple([default] * len(keys))
        return tuple(
            await asyncio.gather(*[self._transform_value(key, value, default) for key, value in zip(keys, values)])
        )

    async def _transform_value(self, key: Key, value: bytes | None, default: Value | None):
        if value is None:
            return default
        if value.isdigit():
            return int(value)
        return await self._serializer.decode(self, key=key, value=value, default=default)

    async def incr(self, key: Key, value: int = 1, expire: float | None = None) -> int:
        if not expire:
            return await self._client.incr(key, amount=value)
        if self._sha.get("INCR_EXPIRE") is None:
            self._sha["INCR_EXPIRE"] = self._client.register_script(_INCR_EXPIRE.replace("\n", " "))
        expire = expire or 0
        expire = int(expire * 1000)
        return await self._sha["INCR_EXPIRE"](keys=(key,), args=(value, expire))

    async def get_bits(self, key: Key, *indexes: int, size: int = 1) -> tuple[int, ...]:
        """
        https://redis.io/commands/bitfield
        """
        bitops = self._client.bitfield(key)
        for index in indexes:
            bitops.get(fmt=f"u{size}", offset=f"#{index}")  # type: ignore[attr-defined]
        return tuple(await bitops.execute() or [])  # type: ignore[attr-defined]

    async def incr_bits(self, key: Key, *indexes: int, size: int = 1, by: int = 1) -> tuple[int, ...]:
        bitops = self._client.bitfield(key)
        for index in indexes:
            bitops.incrby(  # type: ignore[attr-defined]
                fmt=f"u{size}", offset=f"#{index}", increment=by, overflow="SAT"
            )
        return tuple(await bitops.execute())  # type: ignore[attr-defined]

    async def ping(self, message: bytes | None = None) -> bytes:
        await self._client.ping()
        if message is None or message == b"PING":
            return b"PONG"
        return message

    async def set_raw(self, key: Key, value: Value, **kwargs: Any):
        return await self._client.set(key, value, **kwargs)

    async def get_raw(self, key: Key) -> Value:
        return await self._client.get(key)

    async def slice_incr(
        self,
        key: Key,
        start: int | float,
        end: int | float,
        maxvalue: int,
        expire: float | None = None,
    ) -> int:
        expire = expire or 0
        expire = int(expire * 1000)
        if self._sha.get("INCR_SLICE") is None:
            self._sha["INCR_SLICE"] = self._client.register_script(_INCR_SLICE.replace("\n", " "))
        return await self._sha["INCR_SLICE"](keys=(key,), args=(start, end, maxvalue, expire))

    async def set_add(self, key: Key, *values: str, expire: float | None = None):
        if expire is None:
            return await self._client.sadd(key, *values)
        expire = int(expire * 1000)
        async with self._pipeline as pipe:
            await pipe.sadd(key, *values)
            await pipe.pexpire(key, expire)
            await pipe.execute()

    async def set_remove(self, key: Key, *values: str):
        await self._client.srem(key, *values)

    async def set_pop(self, key: Key, count: int = 100) -> Iterable[str]:
        values = await self._client.spop(key, count)
        if values is None:
            return []

        if isinstance(values, bytes):
            values = [values]

        return [value.decode() for value in values]  # type: ignore[union-attr]

    async def get_keys_count(self) -> int:
        return await self._client.dbsize()

    async def close(self):
        if self.__is_init and self._client:
            await self._client.close()
            await self._client.connection_pool.disconnect()
        self.__is_init = False



from __future__ import annotations

from collections.abc import Iterator
from contextlib import contextmanager
from contextvars import ContextVar
from functools import wraps
from typing import Any

from ._typing import AsyncCallable_T
from .backends.interface import _BackendInterface
from .commands import RETRIEVE_CMDS, Command
from .formatter import default_format
from .key import get_call_values
from .key_context import context as template_context


def invalidate(
    backend: _BackendInterface,
    key_template: str,
    args_map: dict[str, str] | None = None,
    defaults: dict[str, Any] | None = None,
):
    args_map = args_map or {}
    defaults = defaults or {}

    def _decor(func: AsyncCallable_T) -> AsyncCallable_T:
        @wraps(func)
        async def _wrap(*args, **kwargs):
            result = await func(*args, **kwargs)
            _args = get_call_values(func, args, kwargs)
            _args.update(defaults)
            for source, dest in args_map.items():
                if dest in _args:
                    _args[source] = _args.pop(dest)
            key = default_format(key_template, **_args)
            with template_context(**_args, rewrite=True):
                await backend.delete_match(key)
            return result

        return _wrap

    return _decor


_INVALIDATE_FURTHER = ContextVar("invalidate", default=False)


@contextmanager
def invalidate_further() -> Iterator[None]:
    _INVALIDATE_FURTHER.set(True)
    try:
        yield
    finally:
        _INVALIDATE_FURTHER.set(False)


async def _aiter(num=0):  # pragma: no cover
    """A trick for typing"""
    for i in range(num):
        yield i


async def _invalidate_middleware(call, cmd: Command, backend: _BackendInterface, *args, **kwargs):
    if _INVALIDATE_FURTHER.get() and cmd in RETRIEVE_CMDS:
        if "key" in kwargs:
            await backend.delete(kwargs["key"])
            return kwargs.get("default")
        if cmd == Command.GET_MATCH:
            await backend.delete_match(kwargs["pattern"])
            return _aiter()
        if cmd == Command.GET_MANY:
            await backend.delete_many(*args)
            return ()
    return await call(*args, **kwargs)



from __future__ import annotations

import time
from contextvars import ContextVar
from functools import wraps
from typing import TYPE_CHECKING, Any, Callable

if TYPE_CHECKING:  # pragma: no cover
    from cashews._typing import CallableCacheCondition, DecoratedFunc

_spent = ContextVar("spent", default=0.0)


def create_time_condition(
    limit: float,
) -> tuple[CallableCacheCondition, Callable[[DecoratedFunc], DecoratedFunc]]:
    def decorator(func: DecoratedFunc) -> DecoratedFunc:
        @wraps(func)
        async def _wrapper(*args, **kwargs):
            start = time.perf_counter()
            try:
                return await func(*args, **kwargs)
            finally:
                _spent.set(time.perf_counter() - start)

        return _wrapper  # type: ignore[return-value]

    def condition(result: Any, args: tuple, kwargs: dict[str, Any], key: str = "") -> bool:
        return _spent.get() > limit

    return condition, decorator



from __future__ import annotations

from datetime import timedelta

from cashews._typing import TTL


def ttl_to_seconds(ttl: TTL, *args, with_callable: bool = False, result=None, **kwargs) -> int | float | None:
    if ttl is None:
        return None
    _type = type(ttl)  # isinstance is slow
    if _type is str:
        return _ttl_from_str(ttl)  # type: ignore[arg-type]
    if _type is int:
        return ttl  # type: ignore[return-value]
    if _type is timedelta:
        return ttl.total_seconds()  # type: ignore[union-attr]

    if callable(ttl) and with_callable:
        try:
            ttl = ttl(*args, result=result, **kwargs)
        except TypeError:
            ttl = ttl(*args, **kwargs)  # type: ignore[operator, misc]
        return ttl_to_seconds(ttl)
    return ttl  # type: ignore[return-value]


_STR_TO_DELTA = {
    "h": timedelta(hours=1),
    "m": timedelta(minutes=1),
    "s": timedelta(seconds=1),
    "d": timedelta(days=1),
}


def _ttl_from_str(ttl: str) -> int:
    result = 0
    mul = ""
    for char in ttl.strip().lower():
        if char.isdigit():
            mul += char
        elif char in _STR_TO_DELTA:
            result += int(mul) * int(_STR_TO_DELTA[char].total_seconds())
            mul = ""
        else:
            raise ValueError(f"ttl '{ttl}' has wrong string representation")
    if mul != "" and not result:
        return int(mul)
    return result



from __future__ import annotations

import logging
from datetime import datetime, timezone
from functools import wraps
from typing import TYPE_CHECKING, Any, Callable, NoReturn

from cashews.backends.interface import _BackendInterface
from cashews.exceptions import RateLimitError
from cashews.key import get_cache_key, get_cache_key_template
from cashews.ttl import ttl_to_seconds

if TYPE_CHECKING:  # pragma: no cover
    from cashews._typing import TTL, DecoratedFunc, KeyOrTemplate

logger = logging.getLogger(__name__)


def _default_action(*args: Any, **kwargs: Any) -> NoReturn:
    raise RateLimitError()


def slice_rate_limit(
    backend: _BackendInterface,
    limit: int,
    period: TTL,
    key: KeyOrTemplate | None = None,
    action: Callable | None = _default_action,
    prefix: str = "srate",
) -> Callable[[DecoratedFunc], DecoratedFunc]:  # pylint: disable=too-many-arguments
    """
    Rate limit for function call. Do not call function if rate limit is reached, and call given action

    :param backend: cache backend
    :param limit: number of calls
    :param period: Period
    :param key: a rate-limit key template
    :param action: call when rate limit reached, default raise RateLimitError
    :param prefix: custom prefix for key, default 'rate_limit'
    """
    period = ttl_to_seconds(period)
    action = action or _default_action

    def decorator(func: DecoratedFunc) -> DecoratedFunc:
        _key_template = get_cache_key_template(func, key=key, prefix=prefix)

        @wraps(func)
        async def wrapped_func(*args, **kwargs):
            _period = ttl_to_seconds(period, *args, **kwargs, with_callable=True)
            _cache_key = get_cache_key(func, _key_template, args, kwargs)

            requests_count = await _get_requests_count(backend, _cache_key, limit, _period)
            if requests_count and requests_count > limit:
                logger.info("Rate limit reach for %s", _cache_key)
                return action(*args, **kwargs)
            return await func(*args, **kwargs)

        return wrapped_func  # type: ignore[return-value]

    return decorator


async def _get_requests_count(backend: _BackendInterface, key: str, limit: int, period: int) -> int:
    timestamp = datetime.now(timezone.utc).timestamp()
    return await backend.slice_incr(key, timestamp - period, timestamp, maxvalue=limit + 1, expire=period)



try:
    from ._bitarray_lib import Bitarray
except ImportError:
    from ._bitarray import Bitarray  # type: ignore[assignment]
from .object_size import get_obj_size
from .split_hash import get_indexes

__all__ = [
    "Bitarray",
    "get_obj_size",
    "get_indexes",
]



from unittest.mock import Mock

from cashews.commands import Command
from cashews.wrapper import Cache


async def test_with_callback_on_set(cache: Cache):
    call = Mock()

    with cache.callback(callback=call, cmd=Command.SET):
        await cache.set("test", "value")

    call.assert_called_once_with("test", result=True)


async def test_with_callback_on_set_skip_get(cache: Cache):
    call = Mock()

    with cache.callback(callback=call, cmd=Command.SET):
        await cache.get(
            "test",
        )

    call.assert_not_called()



from .cache_condition import NOT_NONE, only_exceptions, with_exceptions
from .commands import Command
from .contrib import *  # noqa
from .decorators import CacheDetect, context_cache_detect, fast_condition, thunder_protection
from .exceptions import CacheBackendInteractionError, CircuitBreakerOpen, LockedError, RateLimitError
from .formatter import default_formatter
from .helpers import add_prefix, all_keys_lower, memory_limit
from .key import get_cache_key_template, noself
from .key_context import context as key_context
from .key_context import register as register_key_context
from .validation import invalidate_further
from .wrapper import Cache, TransactionMode, register_backend

# pylint: disable=invalid-name
cache = Cache(name="default")
failover = cache.failover
early = cache.early
soft = cache.soft
hit = cache.hit
transaction = cache.transaction
setup = cache.setup
cache_detect = cache.detect

circuit_breaker = cache.circuit_breaker
dynamic = cache.dynamic
rate_limit = cache.rate_limit
slice_rate_limit = cache.slice_rate_limit
locked = cache.locked

invalidate = cache.invalidate

mem = Cache(name="mem")
mem.setup(
    "mem://?check_interval=1",
    size=1_000_000,
)  # 1_000_000 * 248(small dict size) == 31 mb


__all__ = [
    "mem",
    "cache",
    "failover",
    "early",
    "soft",
    "hit",
    "transaction",
    "setup",
    "cache_detect",
    "circuit_breaker",
    "dynamic",
    "rate_limit",
    "slice_rate_limit",
    "locked",
    "invalidate",
    "NOT_NONE",
    "only_exceptions",
    "with_exceptions",
    "Command",
    "CacheDetect",
    "context_cache_detect",
    "fast_condition",
    "thunder_protection",
    "CacheBackendInteractionError",
    "CircuitBreakerOpen",
    "LockedError",
    "RateLimitError",
    "default_formatter",
    "add_prefix",
    "all_keys_lower",
    "memory_limit",
    "get_cache_key_template",
    "noself",
    "invalidate_further",
    "Cache",
    "TransactionMode",
    "register_backend",
    "key_context",
    "register_key_context",
]



import json
import pickle
from enum import Enum

from ._typing import Value
from .exceptions import UnsupportedPicklerError

_SQLALC_PICKLE = True
try:
    from sqlalchemy.ext import serializer as sqlalchemy_pickle
except ImportError:
    _SQLALC_PICKLE = False
    sqlalchemy_pickle = pickle  # type: ignore[misc,unused-ignore]

_DILL_PICKLE = True
try:
    import dill
except ImportError:
    _DILL_PICKLE = False
    dill = pickle


class Pickler:
    PickleError = pickle.PickleError
    UnpicklingError = (pickle.UnpicklingError, TypeError, ModuleNotFoundError, ImportError)

    @staticmethod
    def loads(value: bytes) -> Value:
        return pickle.loads(value, fix_imports=False, encoding="bytes")

    @staticmethod
    def dumps(value: Value) -> bytes:
        return pickle.dumps(value, protocol=pickle.HIGHEST_PROTOCOL, fix_imports=False)


class SQLAlchemyPickler(Pickler):
    @staticmethod
    def loads(value: bytes) -> Value:
        return sqlalchemy_pickle.loads(
            value,
        )

    @staticmethod
    def dumps(value: Value) -> bytes:
        return sqlalchemy_pickle.dumps(value)


class DillPickler(Pickler):
    @staticmethod
    def loads(value: bytes) -> Value:
        return dill.loads(value)

    @staticmethod
    def dumps(value: Value) -> bytes:
        return dill.dumps(value)


class NonPickler(Pickler):
    @staticmethod
    def loads(value: bytes) -> Value:
        return value

    @staticmethod
    def dumps(value: Value) -> bytes:
        return value


class JsonPickler(Pickler):
    json_serial = None

    @staticmethod
    def loads(value: bytes):
        return json.loads(value)

    @classmethod
    def dumps(cls, value) -> bytes:
        return json.dumps(value, default=cls.json_serial).encode()


class PicklerType(Enum):
    DEFAULT = "default"
    NULL = "null"
    JSON = "json"
    DILL = "dill"
    SQLALCHEMY = "sqlalchemy"


_picklers = {
    PicklerType.DEFAULT: Pickler,
    PicklerType.SQLALCHEMY: SQLAlchemyPickler,
    PicklerType.DILL: DillPickler,
    PicklerType.NULL: NonPickler,
    PicklerType.JSON: JsonPickler,
}


def get_pickler(pickler_type: PicklerType):
    if pickler_type not in _picklers:
        raise UnsupportedPicklerError()

    if pickler_type == PicklerType.SQLALCHEMY and not _SQLALC_PICKLE:
        raise UnsupportedPicklerError()

    if pickler_type == PicklerType.DILL and not _DILL_PICKLE:
        raise UnsupportedPicklerError()

    return _picklers[pickler_type]


DEFAULT_PICKLER = get_pickler(PicklerType.DEFAULT)
